# pylint:disable=isinstance-second-argument-not-valid-type
from __future__ import annotations
from typing import Any
import logging
import time
import contextlib

import claripy
import angr.ailment as ailment
import pyvex

from angr.code_location import CodeLocation
from angr.analyses import ForwardAnalysis, visitors
from angr.knowledge_plugins.propagations.propagation_model import PropagationModel
from angr.knowledge_plugins.propagations.states import PropagatorVEXState, PropagatorState
from angr import sim_options
from angr.analyses import register_analysis
from angr.analyses.analysis import Analysis
from .engine_vex import SimEnginePropagatorVEX

_l = logging.getLogger(name=__name__)


class PropagatorAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=abstract-method
    """
    PropagatorAnalysis implements copy propagation. It propagates values (either constant values or variables) and
    expressions inside a block or across a function.

    PropagatorAnalysis only supports VEX. For AIL, please use SPropagator.

    PropagatorAnalysis performs certain arithmetic operations between constants, including but are not limited to:

    - addition
    - subtraction
    - multiplication
    - division
    - xor

    It also performs the following memory operations:

    - Loading values from a known address
    - Writing values to a stack variable
    """

    def __init__(
        self,
        func=None,
        block=None,
        func_graph=None,
        base_state=None,
        max_iterations=30,
        load_callback=None,
        stack_pointer_tracker=None,
        only_consts=False,
        completed_funcs=None,
        do_binops=True,
        store_tops=True,
        vex_cross_insn_opt=False,
        func_addr: int | None = None,
        gp: int | None = None,
        cache_results: bool = False,
        key_prefix: str | None = None,
        profiling: bool = False,
    ):
        if block is None and func is not None:
            # only func is specified. traversing a function
            self.flavor = "function"
        elif block is not None:
            # traversing a block (but func might be specified at the same time to provide extra information, e.g., the
            # value for register t9 for MIPS32/64 binaries)
            self.flavor = "block"
        else:
            raise ValueError("Unsupported analysis target.")

        start = time.perf_counter_ns() / 1000000 if profiling else 0

        self._base_state = base_state
        self._function = func
        self._func_addr = func_addr if func_addr is not None else (None if func is None else func.addr)
        self._block = block
        self._block_addr = block.addr if block is not None else None
        self._max_iterations = max_iterations
        self._load_callback = load_callback
        self._stack_pointer_tracker = stack_pointer_tracker  # only used when analyzing AIL functions
        self._only_consts = only_consts
        self._completed_funcs = completed_funcs
        self._do_binops = do_binops
        self._store_tops = store_tops
        self._vex_cross_insn_opt = vex_cross_insn_opt
        self._gp = gp
        self._prop_key_prefix = key_prefix
        self._cache_results = cache_results
        self._initial_codeloc: CodeLocation
        self.stmts_to_remove: set[CodeLocation] = set()
        if self.flavor == "function":
            self._initial_codeloc = CodeLocation(self._func_addr, stmt_idx=0, ins_addr=self._func_addr)
        else:  # flavor == "block"
            self._initial_codeloc = CodeLocation(self._block_addr, stmt_idx=0, ins_addr=self._block_addr)

        self.model: PropagationModel = None

        if self._cache_results:
            # Resume the analysis from the previously unfinished result
            self.model = self.kb.propagations.get(self.prop_key, None)

        if self.model is None:
            self.model = PropagationModel(
                self.prop_key,
                function=self._function,
            )
            cache_used = False
        else:
            cache_used = True

        graph_visitor: visitors.SingleNodeGraphVisitor | visitors.FunctionGraphVisitor
        if self.flavor == "block":
            graph_visitor = None
            if self._cache_results:
                graph_visitor: visitors.SingleNodeGraphVisitor | None = self.model.graph_visitor

            if graph_visitor is None:
                graph_visitor = visitors.SingleNodeGraphVisitor(block)

        elif self.flavor == "function":
            graph_visitor = None
            if self._cache_results:
                graph_visitor: visitors.FunctionGraphVisitor | None = self.model.graph_visitor
                if graph_visitor is not None:
                    # resume
                    resumed = graph_visitor.resume_with_new_graph(func_graph if func_graph is not None else func.graph)
                    if not resumed:
                        # clean up...
                        self.model = PropagationModel(self.prop_key, function=self._function)

            if graph_visitor is None:
                graph_visitor = visitors.FunctionGraphVisitor(func, func_graph)
                self.model.graph_visitor = graph_visitor

        else:
            raise TypeError(f"Unsupported flavor {self.flavor}")

        ForwardAnalysis.__init__(
            self, order_jobs=True, allow_merging=True, allow_widening=False, graph_visitor=graph_visitor
        )

        bp_as_gpr = False
        the_func = None
        if self._function is not None:
            the_func = self._function
        else:
            if self._func_addr is not None:
                with contextlib.suppress(KeyError):
                    the_func = self.kb.functions.get_by_addr(self._func_addr)
        if the_func is not None:
            bp_as_gpr = the_func.info.get("bp_as_gpr", False)

        # pyright says pylint is wrong about this
        self._engine_vex = SimEnginePropagatorVEX(  # pylint: disable=abstract-class-instantiated
            project=self.project,
            bp_as_gpr=bp_as_gpr,
        )

        # optimization: skip state copying for the initial state
        self._initial_state = None

        # performance counters
        self._analyzed_states: int = 0

        self._analyze()

        if self._cache_results:
            # update the cache
            self.kb.propagations.update(self.prop_key, self.model)

        if profiling:
            elapsed = time.perf_counter_ns() / 1000000 - start
            if self.flavor == "function":
                _l.warning("%r:", self._function)
                _l.warning("  Blocks: %d", len(self._function.block_addrs_set))
            else:
                _l.warning("%r:", self._block)
            _l.warning("  Time elapsed: %0.02f milliseconds", elapsed)
            _l.warning("  Cache used: %s", cache_used)
            _l.warning("  Analyzed states: %d", self._analyzed_states)

    @property
    def prop_key(self) -> tuple[str | None, str, int, bool, bool, bool]:
        """
        Gets a key that represents the function and the "flavor" of the propagation result.
        """
        addr = self._func_addr if self._func_addr is not None else self._block_addr
        return self._prop_key_prefix, self.flavor, addr, self._do_binops, self._only_consts, self._vex_cross_insn_opt

    @property
    def replacements(self):
        return self.model.replacements

    @replacements.setter
    def replacements(self, v):
        self.model.replacements = v

    #
    # Main analysis routines
    #

    def _node_key(self, node: ailment.Block | pyvex.IRSB) -> Any:
        if type(node) is ailment.Block:
            return node.addr, node.idx
        if type(node) is pyvex.IRSB:
            return node.addr
        # fallback
        return node

    def _pre_analysis(self):
        pass

    def _pre_job_handling(self, job):
        pass

    def _initial_abstract_state(self, node):
        self._initial_state = PropagatorVEXState.initial_state(
            self.project,
            only_consts=self._only_consts,
            gp=self._gp,
            do_binops=self._do_binops,
            store_tops=self._store_tops,
            func_addr=self._func_addr,
            max_prop_expr_occurrence=1 if self.flavor == "function" else 0,
            initial_codeloc=self._initial_codeloc,
            model=self.model,
        )
        return self._initial_state

    def _merge_states(self, node, *states: PropagatorState):
        merged_state, merge_occurred = states[0].merge(*states[1:])
        return merged_state, not merge_occurred

    def _run_on_node(self, node, state):
        self._analyzed_states += 1

        block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=self._vex_cross_insn_opt)
        block_key = node.addr
        engine = self._engine_vex
        if block.size == 0:
            # maybe the block is not decodeable
            return False, state

        if state is not self._initial_state:
            # make a copy of the state if it's not the initial state
            state = state.copy()
            state.init_replacements()
        else:
            # clear self._initial_state so that we *do not* run this optimization again!
            self._initial_state = None

        # Suppress spurious output
        if self._base_state is not None:
            self._base_state.options.add(sim_options.SYMBOL_FILL_UNCONSTRAINED_REGISTERS)
            self._base_state.options.add(sim_options.SYMBOL_FILL_UNCONSTRAINED_MEMORY)

        self.model.input_states[block_key] = state.copy()

        state = engine.process(
            state,
            block=block,
            project=self.project,
            base_state=self._base_state,
            load_callback=self._load_callback,
            fail_fast=self._fail_fast,
        )
        state.filter_replacements()

        self.model.node_iterations[block_key] += 1
        self.model.states[block_key] = state
        self.model.block_initial_reg_values.update(state.block_initial_reg_values)

        if self.model.replacements is None:
            self.model.replacements = state._replacements
        else:
            PropagatorState.merge_replacements(self.model.replacements, state._replacements)

        # TODO: Clear registers according to calling conventions

        if self.model.node_iterations[block_key] < self._max_iterations:
            return None, state
        return False, state

    def _process_input_state_for_successor(self, node, successor, input_state: PropagatorVEXState):
        if self._only_consts and isinstance(input_state, PropagatorVEXState):
            key = node.addr, successor.addr
            if key in self.model.block_initial_reg_values:
                input_state: PropagatorVEXState = input_state.copy()
                for reg_offset, reg_size, value in self.model.block_initial_reg_values[key]:
                    input_state.store_register(reg_offset, reg_size, claripy.BVV(value, reg_size * 8))
        return input_state

    def _intra_analysis(self):
        pass

    def _check_func_complete(self, func):
        """
        Checks if a function is completely created by the CFG. Completed
        functions are passed to the Propagator at initialization. Defaults to
        being empty if no pass is initiated.

        :param func:    Function to check (knowledge_plugins.functions.function.Function)
        :return:        Bool
        """
        complete = False
        if self._completed_funcs is None:
            return complete

        if func.addr in self._completed_funcs:
            complete = True

        return complete

    def _post_analysis(self):
        """
        Post Analysis of Propagation().
        We add the current propagation replacements result to the kb if the
        function has already been completed in cfg creation.
        """

        # Filter replacements and remove all TOP values
        if self.model.replacements is not None:
            for codeloc in list(self.model.replacements.keys()):
                filtered_rep = {}
                for k, v in self.model.replacements[codeloc].items():
                    if isinstance(v, claripy.ast.Base) and not PropagatorState.is_top(v):
                        filtered_rep[k] = v
                self.model.replacements[codeloc] = filtered_rep

        if self._cache_results:
            self.kb.propagations.update(self.prop_key, self.model)

    def _analyze(self):
        """
        The main analysis for Propagator. Overwritten to include an optimization to stop
        analysis if we have already analyzed the entire function once.
        """
        self._pre_analysis()

        # normal analysis execution
        if self._graph_visitor is None:
            # There is no base graph that we can rely on. The analysis itself should generate successors for the
            # current job.
            # An example is the CFG recovery.

            self._analysis_core_baremetal()

        else:
            # We have a base graph to follow. Just handle the current job.

            self._analysis_core_graph()

        self._post_analysis()


register_analysis(PropagatorAnalysis, "Propagator")



from __future__ import annotations

import logging

from .base import ExplorationTechnique

l = logging.getLogger(name=__name__)


class ManualMergepoint(ExplorationTechnique):
    def __init__(self, address, wait_counter=10, prune=True):
        super().__init__()
        self.address = address
        self.wait_counter_limit = wait_counter
        self.prune = prune
        self.wait_counter = 0
        self.stash = f"merge_waiting_{self.address:#x}_{id(self):x}"
        self.filter_marker = f"skip_next_filter_{self.address:#x}"

    def setup(self, simgr):
        simgr.stashes[self.stash] = []

    def mark_nofilter(self, simgr, stash):
        for state in simgr.stashes[stash]:
            state.globals[self.filter_marker] = True

    def mark_okfilter(self, simgr, stash):
        for state in simgr.stashes[stash]:
            state.globals.pop(self.filter_marker)

    def step(self, simgr, stash="active", **kwargs):
        # ha ha, very funny, if this is being run on a single-step basis our filter probably misfired
        if len(simgr.stashes[self.stash]) == 1 and len(simgr.stashes[stash]) == 0:
            simgr = simgr.move(self.stash, stash)

        # perform all our analysis as a post-mortem on a given step
        stop_points = kwargs.pop("extra_stop_points", set())
        stop_points.add(self.address)
        simgr = simgr.step(stash=stash, extra_stop_points=stop_points, **kwargs)
        # self.mark_okfilter(simgr, stash)

        # do filtering
        new_stash = []
        for state in simgr.stashes[stash]:
            if self.filter_marker not in state.globals and state.addr == self.address:
                self.wait_counter = 0
                simgr.stashes[self.stash].append(state)
            else:
                new_stash.append(state)
        simgr.stashes[stash][:] = new_stash

        # nothing to do if there's no states waiting
        if len(simgr.stashes[self.stash]) == 0:
            return simgr

        # tick the counter
        self.wait_counter += 1

        # see if it's time to merge (out of active or hit the wait limit)
        if len(simgr.stashes[stash]) != 0 and self.wait_counter < self.wait_counter_limit:
            return simgr

        # self.mark_nofilter(simgr, self.stash)

        # only both merging if, you know, there's actually states to merge
        if len(simgr.stashes[self.stash]) == 1:
            simgr.move(self.stash, stash)
            return simgr

        # do the merge, keyed by unique callstack
        l.info("Merging %d states at %#x", len(simgr.stashes[self.stash]), self.address)
        num_unique = 0
        while len(simgr.stashes[self.stash]):
            num_unique += 1
            exemplar_callstack = simgr.stashes[self.stash][0].callstack
            simgr.move(self.stash, "merge_tmp", lambda s, ec=exemplar_callstack: s.callstack == ec)
            l.debug("...%d with unique callstack #%d", len(simgr.merge_tmp), num_unique)
            if len(simgr.merge_tmp) > 1:
                simgr = simgr.merge(stash="merge_tmp", prune=self.prune)
            simgr = simgr.move("merge_tmp", stash)

        return simgr



from __future__ import annotations

from .base import SimConcretizationStrategy


class SimConcretizationStrategyAny(SimConcretizationStrategy):
    """
    Concretization strategy that returns any single solution.
    """

    def _concretize(self, memory, addr, **kwargs):
        if self._exact:
            return [self._any(memory, addr, **kwargs)]
        mn, mx = self._range(memory, addr, **kwargs)
        if mn == mx:
            return [mn]
        return None



from __future__ import annotations
from typing import TYPE_CHECKING

from angr.ailment import Block
from angr.ailment.statement import Label, ConditionalJump
from angr.ailment.block_walker import AILBlockViewer

from angr.analyses.decompiler.sequence_walker import SequenceWalker

if TYPE_CHECKING:
    from angr.ailment import Address
    from angr.ailment.statement import Call, SideEffectStatement


class AILBlockCallCounter(AILBlockViewer):
    """
    Helper class to count AIL calls and call-expressions in a block, or collect call statements and call expressions
    as well as their locations.
    """

    def __init__(self, consider_conditions: bool = False):
        super().__init__()
        self.calls: int = 0
        self.consider_conditions = consider_conditions
        self.call_stmts: list[tuple[tuple[Address | None, int], SideEffectStatement]] = []
        self.call_exprs: list[tuple[tuple[Address | None, int], Call]] = []

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Block | None):
        if not self.consider_conditions:
            return
        super()._handle_ConditionalJump(stmt_idx, stmt, block)

    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt, block: Block | None):
        self.calls += 1
        self.call_exprs.append((((block.addr, block.idx) if block is not None else None, stmt_idx), expr))
        super()._handle_CallExpr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_SideEffectStatement(self, stmt_idx: int, stmt: SideEffectStatement, block: Block | None):
        self.call_stmts.append((((block.addr, block.idx) if block is not None else None, stmt_idx), stmt))
        super()._handle_SideEffectStatement(stmt_idx, stmt, block)


class AILCallCounter(SequenceWalker):
    """
    Helper class to count AIL calls and call expressions in a structuring node, or collect call statements and call
    expressions as well as their locations.
    """

    def __init__(self, consider_conditions: bool = False):
        handlers = {
            Block: self._handle_Block,
        }
        super().__init__(handlers)
        self.calls = 0
        self.non_label_stmts = 0
        self.consider_conditions = consider_conditions
        self.call_stmts: list[tuple[tuple[Address | None, int], Call]] = []
        self.call_exprs: list[tuple[tuple[Address | None, int], Call]] = []

    def _handle_Condition(self, node, **kwargs):
        if self.consider_conditions:
            super()._handle(node, **kwargs)
        else:
            # do not count calls in conditions
            if node.true_node is not None:
                super()._handle(node.true_node, **kwargs)
            if node.false_node is not None:
                super()._handle(node.false_node, **kwargs)

    def _handle_Block(self, node: Block, **kwargs):  # pylint:disable=unused-argument
        ctr = AILBlockCallCounter(consider_conditions=self.consider_conditions)
        ctr.walk(node)
        self.calls += ctr.calls
        self.call_stmts += ctr.call_stmts
        self.call_exprs += ctr.call_exprs
        self.non_label_stmts += sum(1 for stmt in node.statements if not isinstance(stmt, Label))



from __future__ import annotations
import claripy

from angr.storage.memory_mixins.memory_mixin import MemoryMixin
from angr.storage.memory_mixins.paged_memory.pages.ispo_mixin import ISPOMixin
from angr.errors import SimMergeError


class SlottedMemoryMixin(MemoryMixin):
    def __init__(self, width=None, **kwargs):
        super().__init__(**kwargs)

        if width is None and isinstance(self, ISPOMixin):
            width = 4
        self.width = width

        self.contents = {}

    def set_state(self, state):
        super().set_state(state)
        if self.width is None:
            self.width = state.arch.bytes

    def copy(self, memo):
        o = super().copy(memo)
        o.width = self.width
        o.contents = dict(self.contents)
        return o

    def merge(self, others, merge_conditions, common_ancestor=None):
        if any(o.width != self.width for o in others):
            raise SimMergeError("Cannot merge slotted memory with disparate widths")
        addr_set = set(self.contents)
        for o in others:
            addr_set.update(o.contents)

        for addr in addr_set:
            self._single_store(
                addr,
                0,
                self.width,
                claripy.ite_cases(
                    zip(merge_conditions[1:], (o._single_load(addr, 0, self.width) for o in others)),
                    self._single_load(addr, 0, self.width),
                ),
            )
        # FIXME: Return True only when merge actually happens
        return True

    def _resolve_access(self, addr, size):
        """
        Resolves a memory access of a certain size. Returns a sequence of the bases, offsets, and sizes of the accesses
        required to fulfil this.
        """

        # if we fit in one word
        first_offset = addr % self.width
        first_base = addr - first_offset
        if first_offset + size <= self.width:
            result = [(first_base, first_offset, size)]
        else:
            last_size = (addr + size) % self.width
            last_base = addr + size - last_size

            result = [(first_base, first_offset, self.width - first_offset)]
            result.extend((a, 0, self.width) for a in range(first_base + self.width, last_base, self.width))
            if last_size != 0:
                result.append((last_base, 0, last_size))

        # little endian: need to slice in reverse and also concatenate in reverse
        # if load endness and storage endness don't match, reverse whole data value before store/after load
        if self.endness == "Iend_LE":
            result = [(addr, self.width - offset - size, size) for addr, offset, size in reversed(result)]

        return result

    def _single_load(self, addr, offset, size, **kwargs):
        """
        Performs a single load.
        """
        try:
            d = self.contents[addr]
        except KeyError:
            d = self._default_value(addr, self.width, key=(*self.variable_key_prefix, addr), **kwargs)
            self.contents[addr] = d

        if offset == 0 and size == self.width:
            return d
        return d.get_bytes(offset, size)

    def _single_store(self, addr, offset, size, data):
        """
        Performs a single store.
        """

        if offset == 0 and size == self.width:
            self.contents[addr] = data
        elif offset == 0:
            cur = self._single_load(addr, size, self.width - size)
            self.contents[addr] = data.concat(cur)
        elif offset + size == self.width:
            cur = self._single_load(addr, 0, offset)
            self.contents[addr] = cur.concat(data)
        else:
            cur = self._single_load(addr, 0, self.width)
            start = cur.get_bytes(0, offset)
            end = cur.get_bytes(offset + size, self.width - offset - size)
            self.contents[addr] = start.concat(data, end)

    def load(self, addr, size=None, *, endness=None, **kwargs):
        accesses = self._resolve_access(addr, size)

        value = claripy.Concat(*(self._single_load(addr, offset, size) for addr, offset, size in accesses))
        if endness != self.endness:
            value = value.reversed

        return value

    def store(self, addr, data, size=None, *, endness=None, **kwargs):
        if endness != self.endness:
            data = data.reversed

        accesses = self._resolve_access(addr, size)
        cur_offset = 0
        for addr, offset, size in accesses:
            piece = data.get_bytes(cur_offset, size)
            self._single_store(addr, offset, size, piece)
            cur_offset += size

    def changed_bytes(self, other):
        changes = set()

        for addr, v in self.contents.items():
            for i in range(self.width):
                other_byte = other.load(addr + i, 1)
                our_byte = v.get_byte(i)
                if other_byte is our_byte:
                    changes.add(addr + i)

        return changes



from __future__ import annotations
import logging

import networkx

from angr.analyses.analysis import Analysis, AnalysesHub
from angr.knowledge_plugins.cfg import CFGModel

_l = logging.getLogger(__name__)


class ObfuscationDetector(Analysis):
    """
    This analysis detects, usually in ways that are more robust than section name matching or signature matching, the
    existence of obfuscation techniques in a binary.
    """

    def __init__(self, cfg: CFGModel | None = None):
        self.obfuscated: bool = False
        self.possible_obfuscators: list[str] = []

        if cfg is None:
            _l.warning(
                "PackingDetector is using a most accurate CFG model in the knowledge base. We assume it is "
                "generated with force_smart_scan=False and force_complete_scan=False."
            )
            self._cfg = self.kb.cfgs.get_most_accurate()
        else:
            self._cfg = cfg

        self.analyze()

    def analyze(self):
        analysis_routines = [
            self._analyze_vmprotect,
        ]

        for routine in analysis_routines:
            tool = routine()
            if tool:
                self.obfuscated = True
                self.possible_obfuscators.append(tool)

    def _analyze_vmprotect(self) -> str | None:
        """
        We detect VMProtect v3 (with control-flow obfuscation) based on two main characteristics:

        - In amd64 binaries, there exists a strongly connected component in the call graph with over 1,000 nodes.
          Edge/node ratio is >= 1.3
        - There is a high number of pushf and popf instructions in the visible functions.
        """

        high_scc_node_edge_ratio = False
        high_pushf = False
        high_popf = False
        high_clc = False  # pylint:disable=unused-variable

        if self.project.arch.name == "AMD64":
            cg = self.kb.functions.callgraph
            sccs = networkx.strongly_connected_components(cg)

            for scc in sccs:
                subgraph = networkx.subgraph(cg, scc)
                node_count = len(scc)
                if node_count > 1000:
                    edge_count = len(subgraph.edges)

                    if edge_count / node_count >= 1.3:
                        high_scc_node_edge_ratio = True
                        break
        else:
            high_scc_node_edge_ratio = True

        pushf_ctr = 0
        popf_ctr = 0
        clc_ctr = 0  # only used for x86
        is_x86 = self.project.arch.name == "X86"
        cfg_node_count = len(self._cfg.graph)
        for node in self._cfg.nodes():
            if node.size > 0 and node.instruction_addrs:
                block = node.block
                for insn in block.capstone.insns:
                    if insn.mnemonic in {"pushf", "pushfd", "pushfq"}:
                        pushf_ctr += 1
                    elif insn.mnemonic in {"popf", "popfd", "popfq"}:
                        popf_ctr += 1
                    elif is_x86 and insn.mnemonic == "clc":
                        clc_ctr += 1

        if pushf_ctr > cfg_node_count * 0.002:
            high_pushf = True
        if popf_ctr > cfg_node_count * 0.002:
            high_popf = True
        if not is_x86 or clc_ctr > cfg_node_count * 0.002:
            high_clc = True  # noqa: F841

        if high_scc_node_edge_ratio and high_pushf and high_popf:
            return "vmprotect"
        return None


AnalysesHub.register_default("ObfuscationDetector", ObfuscationDetector)



# pylint:disable=raise-missing-from
from __future__ import annotations

from typing import TypeVar, Generic, cast, TYPE_CHECKING, overload
from collections.abc import Iterator, Generator
from collections import OrderedDict, UserDict
import logging
import collections.abc
import re
import weakref
import bisect
import os
import threading
from collections import defaultdict

import lmdb
import networkx

from archinfo.arch_soot import SootMethodDescriptor
import cle
from cachetools import LRUCache
from sortedcontainers import SortedKeysView, SortedItemsView, SortedValuesView

from angr.errors import SimEngineError
from angr.codenode import FuncNode, HookNode
from angr.knowledge_plugins.plugin import KnowledgeBasePlugin
from angr.utils.smart_cache import SmartLRUCache
from angr.protos import function_pb2
from .function import Function
from .soot_function import SootFunction

K = TypeVar("K", int, SootMethodDescriptor)
T = TypeVar("T")

if TYPE_CHECKING:
    from angr import KnowledgeBase
    from angr.knowledge_plugins.rtdb import RuntimeDb

    class SortedDict(Generic[K, T], dict[K, T]):
        def irange(self, *args, **kwargs) -> Iterator[K]: ...

    class SortedList(Generic[K], list[K]):
        def irange(self, *args, **kwargs) -> Iterator[K]: ...
        def add(self, value: K) -> None: ...

else:
    from sortedcontainers import SortedDict, SortedList

QUERY_PATTERN = re.compile(r"^(::(.+?))?::(.+)$")
ADDR_PATTERN = re.compile(r"^(0x[\dA-Fa-f]+)|(\d+)$")

l = logging.getLogger(name=__name__)
_missing = object()

# a global flag to disable SpillingFunctionDict usage; mainly for testing purposes
USE_SPILLING_FUNCTION_DICT = os.environ.get("USE_SPILLING_FUNCTION_DICT", "True").lower() not in ("0", "false", "no")


class FunctionDictBase(Generic[K]):
    """
    Base class for FunctionDict and SpillingFunctionDict.
    """

    def __init__(self, backref: FunctionManager[K] | None, key_types: type = int):
        self._backref = (
            cast(FunctionManager[K], backref if isinstance(backref, weakref.ProxyType) else weakref.proxy(backref))
            if backref is not None
            else None
        )
        self._key_types = key_types

    def _getitem_create_new_function(self, addr: K) -> Function:
        # Create a new function
        if isinstance(addr, SootMethodDescriptor):
            t = SootFunction(self._backref, addr)
        else:
            t = Function(self._backref, addr)
        self[addr] = t
        if self._backref is not None:
            self._backref._function_added(t)
        return t

    def floor_addr(self, addr: K):
        try:
            return next(self.irange(maximum=addr, reverse=True))
        except StopIteration as err:
            raise KeyError(addr) from err

    def ceiling_addr(self, addr):
        try:
            return next(self.irange(minimum=addr, reverse=False))
        except StopIteration as err:
            raise KeyError(addr) from err

    def __setstate__(self, state):
        for k, v in state.items():
            self[k] = v

    def __getstate__(self):
        return dict(self.items())

    def __setitem__(self, addr: K, value: Function):
        raise NotImplementedError

    def __getitem__(self, addr: K) -> Function:
        raise NotImplementedError

    def __delitem__(self, key: K) -> None:
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    @overload
    def get(self, key: K, default: None = None, /, meta_only: bool = False) -> Function: ...
    @overload
    def get(self, key: K, default: Function, /, meta_only: bool = False) -> Function: ...
    @overload
    def get(self, key: K, default: T, /, meta_only: bool = False) -> Function | T: ...

    def get(self, addr: K, default=_missing, /, meta_only: bool = False):
        raise NotImplementedError

    def irange(self, minimum=None, maximum=None, inclusive=(True, True), reverse=False):
        raise NotImplementedError


class FunctionDict(SortedDict[K, Function], FunctionDictBase[K]):
    """
    FunctionDict is a dict where the keys are function starting addresses and map to the associated :class:`Function`.
    """

    def __init__(self, backref: FunctionManager[K] | None, *args, key_types: type = int, **kwargs):
        SortedDict.__init__(self, *args, **kwargs)
        FunctionDictBase.__init__(self, backref, key_types=key_types)

    def copy(self) -> FunctionDict[K]:
        return FunctionDict(self._backref, self, key_types=self._key_types)

    def __getitem__(self, addr: K) -> Function:
        try:
            return super().__getitem__(addr)
        except KeyError as ex:
            if not isinstance(addr, self._key_types):
                raise TypeError(f"FunctionDict only supports {self._key_types} as key type") from ex
            return self._getitem_create_new_function(addr)

    @overload
    def get(self, key: K, default: None = None, /, meta_only: bool = False) -> Function: ...
    @overload
    def get(self, key: K, default: Function, /, meta_only: bool = False) -> Function: ...  # type: ignore
    @overload
    def get(self, key: K, default: T, /, meta_only: bool = False) -> Function | T: ...  # type: ignore

    def get(self, addr: K, default=_missing, /, meta_only: bool = False):  # type: ignore #pylint:disable=unused-argument
        try:
            return super().__getitem__(addr)
        except KeyError:
            pass

        if default is _missing:
            raise KeyError(addr)
        return default

    def floor_addr(self, addr: K):
        try:
            return next(self.irange(maximum=addr, reverse=True))
        except StopIteration as err:
            raise KeyError(addr) from err

    def ceiling_addr(self, addr):
        try:
            return next(self.irange(minimum=addr, reverse=False))
        except StopIteration as err:
            raise KeyError(addr) from err

    def __setstate__(self, state):
        for v, k in state.items():
            self[k] = v

    def __getstate__(self):
        return dict(self.items())


class SpillingFunctionDict(UserDict[K, Function], FunctionDictBase[K]):
    """
    SpillingFunctionDict extends FunctionDict with LRU caching and LMDB spilling. This class keeps only the most
    recently accessed N functions in memory, spilling others to an LMDB database on disk.

    SpillingFunctionDict also keeps a cache of meta-only Function objects that do not have graph or block information.
    These meta-only Function objects are read-only and may become stale if the full Function is later loaded and
    updated. Therefore, please be extremely cautious when using these meta-only Function objects.

    SpillingFunctionDict._load_from_lmdb() does not support reentry. If a function being loaded from LMDB triggers
    another load from LMDB, the inner load will raise a RuntimeError.

    A Function instance becomes "dirty" (Function.dirty) if it has been modified since being loaded from LMDB. Dirty
    functions are always saved back to LMDB when evicted, while clean functions are skipped during eviction.

    A Function instance becomes "evicted" (Function.evicted) and stale if it has been spilled to LMDB. An evicted
    Function instance may not reflect the most recent state of the function, and any changes to an evicted Function
    instance will not be saved to LMDB. Therefore, if SpillingFunctionDict is in use, it is advised not to hold a
    Function instance for too long before using it.

    :ivar cache_limit:          The maximum number of functions to keep in memory.
    :ivar rtdb:                 A reference to the RuntimeDb knowledge base plugin.
    :ivar _lru_order:           An OrderedDict tracking the eviction order of cached functions.
    :ivar _spilled_keys:        A set of function addresses that have been spilled to LMDB.
    :ivar _db_batch_size:       The number of functions that are evicted in a single batch.
    :ivar _meta_func_cache:     An LRU cache for meta-only Function objects.
    :ivar _eviction_enabled:    A flag indicating whether eviction is currently enabled or not.
    """

    def __init__(
        self,
        backref: FunctionManager[K] | None,
        rtdb: RuntimeDb,
        /,
        key_types: type = int,
        cache_limit: int = 1000,
        db_batch_size: int = 100,
        **kwargs,
    ):
        UserDict.__init__(self, **kwargs)
        FunctionDictBase.__init__(self, backref, key_types=key_types)

        self._cache_limit: int = cache_limit
        self.rtdb: RuntimeDb = rtdb
        self._lru_order: OrderedDict[K, None] = OrderedDict()
        self._spilled_keys: set[K] = set()
        self._list = SortedList()  # a sorted list of all keys (cached + spilled)
        self.irange = self._list.irange

        self._meta_func_cache: LRUCache[K, Function] = SmartLRUCache(
            maxsize=cache_limit, evict=self._meta_func_cache_evicted
        )
        self._funcsdb: str | None = None
        self._db_batch_size: int = db_batch_size
        self._eviction_enabled: bool = True
        self._loading_from_lmdb: bool = False
        self._db_load_lock = threading.Lock()
        self._db_store_lock = threading.Lock()

    def __del__(self):
        self._cleanup_lmdb()

    def __getitem__(self, addr: K) -> Function:
        # First try to get from in-memory cache
        if self.is_cached(addr):
            try:
                func = super().__getitem__(addr)
                # Touch to update LRU order
                self._touch(addr)
                return func
            except KeyError:
                pass

        # not found in memory
        if not isinstance(addr, self._key_types):
            raise TypeError(f"SpillingFunctionDict only supports {self._key_types} as key type")

        # Try to load from LMDB if it's spilled
        if addr in self._spilled_keys:
            func = self._load_from_lmdb(addr)
            if func is not None:
                return func

        return self._getitem_create_new_function(addr)

    def __setitem__(self, key: K, value: Function) -> None:
        if key not in self:
            self._list.add(key)
        super().__setitem__(key, value)
        self._on_function_stored(key)

    def __delitem__(self, key: K) -> None:
        # Remove from in-memory map if present
        if self.is_cached(key):
            super().__delitem__(key)
        # Remove from spilled set if present; don't remove it from lmdb to save time
        if key in self._spilled_keys:
            self._spilled_keys.discard(key)
        # Remove from LRU order
        if key in self._lru_order:
            del self._lru_order[key]
        # Remove from sorted list
        self._list.remove(key)

    def __setstate__(self, state: dict):
        self._cache_limit = state["cache_limit"]
        self._db_batch_size = state["db_batch_size"]
        self.data = {}
        self.rtdb = None  # type: ignore
        self._lru_order = OrderedDict()
        self._spilled_keys = set()
        self._list = SortedList()
        self.irange = self._list.irange

        self._meta_func_cache = LRUCache(maxsize=self._cache_limit)
        self._funcsdb = None
        self._eviction_enabled = True
        self._loading_from_lmdb = False
        self._db_load_lock = threading.Lock()
        self._db_store_lock = threading.Lock()

        for k, v in state["items"].items():
            self[k] = v

    def __getstate__(self):
        return {
            "cache_limit": self._cache_limit,
            "db_batch_size": self._db_batch_size,
            "items": dict(self.items()),
        }

    def copy(self) -> SpillingFunctionDict[K]:
        new_dict = SpillingFunctionDict(
            self._backref,
            self.rtdb,
            key_types=self._key_types,
            cache_limit=self._cache_limit,
        )
        # Temporarily disable eviction during copy
        new_dict._eviction_enabled = False
        # iterate over in-memory functions and copy them
        for address in self.cached_keys:
            function = super().__getitem__(address)
            super(SpillingFunctionDict, new_dict).__setitem__(address, function.copy())
            new_dict._lru_order[address] = None

        # Copy any remaining spilled addresses and their LMDB data
        if self._spilled_keys and self._funcsdb is not None:
            new_dict._init_lmdb()
            with (
                self.rtdb.begin_txn(self._funcsdb) as src_txn,
                self.rtdb.begin_txn(new_dict._funcsdb, write=True) as dst_txn,
            ):
                for addr in self._spilled_keys:
                    key = str(addr).encode("utf-8")
                    value = src_txn.get(key)
                    if value is not None:
                        dst_txn.put(key, value)
                        new_dict._spilled_keys.add(addr)

        new_dict._eviction_enabled = True
        return new_dict

    def clear(self) -> None:
        """
        Clear all functions from memory and spilled storage.
        """
        super().clear()
        self._lru_order.clear()
        self._spilled_keys.clear()
        self._cleanup_lmdb()

    def __contains__(self, item) -> bool:
        # Check both in-memory and spilled
        return super().__contains__(item) or item in self._spilled_keys

    def __len__(self) -> int:
        # Total count includes both in-memory and spilled functions
        return self.cached_count + len(self._spilled_keys)

    def __iter__(self):
        # Iterate over all function addresses (cached + spilled)
        return self._list.__iter__()

    @overload
    def get(self, key: K, default: None = None, /, meta_only: bool = False) -> Function: ...
    @overload
    def get(self, key: K, default: Function, /, meta_only: bool = False) -> Function: ...  # type: ignore
    @overload
    def get(self, key: K, default: T, /, meta_only: bool = False) -> Function | T: ...  # type: ignore

    def get(self, addr, default=_missing, /, meta_only: bool = False):  # type: ignore
        # First check in-memory
        if self.is_cached(addr):
            try:
                func = super().__getitem__(addr)
                self._touch(addr)
                return func
            except KeyError:
                pass

        # Check if spilled to LMDB
        if addr in self._spilled_keys:
            if meta_only and addr in self._meta_func_cache:
                return self._meta_func_cache[addr]
            func = self._load_from_lmdb(addr, meta_only=meta_only)
            if func is not None:
                return func

        if default is _missing:
            raise KeyError(addr)
        return default

    def keys(self):
        return SortedKeysView(self)

    def items(self):
        return SortedItemsView(self)

    def values(self):
        return SortedValuesView(self)

    @staticmethod
    def _meta_func_cache_evicted(addr: K, func: Function) -> None:  # pylint: disable=unused-argument
        """
        Callback when a meta-only function is evicted from the meta function cache.
        """
        func.evicted = True

    #
    # Properties
    #

    @property
    def cached_keys(self) -> Generator[K]:
        yield from self._list

    @property
    def cache_limit(self) -> int:
        """
        Get the maximum number of functions to keep in memory.
        """
        return self._cache_limit

    @cache_limit.setter
    def cache_limit(self, value: int) -> None:
        """
        Set the maximum number of functions to keep in memory.
        """
        self._cache_limit = value
        if self.cached_count > value + self._db_batch_size:
            self._evict_lru()

    @property
    def cached_count(self) -> int:
        """Return the number of functions currently in memory."""
        return super().__len__()

    @property
    def spilled_count(self) -> int:
        """Return the number of functions currently spilled to LMDB."""
        return len(self._spilled_keys)

    @property
    def total_count(self) -> int:
        """Return the total number of functions (in memory + spilled)."""
        return super().__len__() + len(self._spilled_keys)

    def is_cached(self, addr: K) -> bool:
        return super().__contains__(addr)

    #
    # LRU Cache Management
    #

    def _touch(self, addr: K) -> None:
        """
        Update the LRU order for a function (move to end = most recently used).
        """
        if addr in self._lru_order:
            self._lru_order.move_to_end(addr)
        else:
            self._lru_order[addr] = None
        if addr in self._meta_func_cache:
            del self._meta_func_cache[addr]

    def _on_function_stored(self, addr: K) -> None:
        self._touch(addr)

        # Remove from spilled set if it was there
        self._spilled_keys.discard(addr)

        # Check if we need to evict (but not during loading)
        if self._eviction_enabled and self._cache_limit is not None and self.cached_count > self._cache_limit:
            self._evict_lru()

    def _evict_lru(self) -> bool:
        """
        Evict functions until we're back under the cache limit.

        :return: True if at least one function was successfully evicted, False otherwise.
        """
        with self._db_store_lock:
            evicted_any = False
            while self.cached_count > self._cache_limit:
                if self._evict_n(min(self._db_batch_size, self.cached_count)) == 0:
                    break
                evicted_any = True
            return evicted_any

    def _evict_n(self, n: int) -> int:
        """
        Evict n least recently used functions to LMDB.

        :return: The number of functions that were successfully evicted.
        """
        if not self._lru_order:
            return 0

        evicted = 0
        funcs_to_evict = []
        for lru_addr in list(self._lru_order):
            if evicted >= n:
                break

            # Don't evict if it's not in memory
            if not self.is_cached(lru_addr):
                self._lru_order.pop(lru_addr)
                continue

            # Get the function
            func = super().__getitem__(lru_addr)
            func.evicted = True
            if func.dirty:
                funcs_to_evict.append(func)

            # Remove from in-memory map
            super().__delitem__(lru_addr)

            # Remove from LRU order
            del self._lru_order[lru_addr]

            # Add to spilled set
            self._spilled_keys.add(lru_addr)
            evicted += 1

            # l.debug("Evicted function %s", hex(lru_addr) if isinstance(lru_addr, int) else lru_addr)

        if funcs_to_evict:
            self._save_to_lmdb(funcs_to_evict)

        return evicted

    #
    # LMDB Management
    #

    def _init_lmdb(self) -> None:
        """
        Lazily initialize the LMDB database for spilling functions.
        """
        if self._funcsdb is None:
            self._funcsdb = self.rtdb.open_db("functions")
        l.debug("Initialized LRU cache LMDB.")

    def _cleanup_lmdb(self):
        """
        Clean up LMDB resources.
        """
        if self._funcsdb is not None:
            self.rtdb.drop_db(self._funcsdb)
            self._funcsdb = None

    def _save_to_lmdb(self, funcs: list[Function]) -> None:
        """
        Save multiple functions to LMDB.
        """
        self._init_lmdb()
        assert self._funcsdb is not None

        while True:
            try:
                with self.rtdb.begin_txn(self._funcsdb, write=True) as txn:
                    for func in funcs:
                        cmsg = func.serialize_to_cmessage()
                        key = str(func.addr).encode("utf-8")
                        txn.put(key, cmsg.SerializeToString())
                break
            except lmdb.MapFullError:
                # Increase map size and retry
                self.rtdb.increase_lmdb_map_size()

    def _delete_from_lmdb(self, addr: K) -> None:
        """
        Delete a function from LMDB.
        """
        if self._funcsdb is None:
            return

        key = str(addr).encode("utf-8")

        with self.rtdb.begin_txn(self._funcsdb, write=True) as txn:
            txn.delete(key)

    def _load_from_lmdb(self, addr: K, meta_only: bool = False) -> Function | None:
        """
        Load a function from LMDB and bring it back into memory.
        """
        if self._funcsdb is None:
            return None

        with self._db_load_lock:
            return self._load_from_lmdb_core(addr, meta_only=meta_only)

    def _load_from_lmdb_core(self, addr: K, meta_only: bool = False) -> Function | None:
        if self._loading_from_lmdb:
            raise RuntimeError("Recursive loading from LMDB detected. This is a bug.")

        self._loading_from_lmdb = True

        try:
            key = str(addr).encode("utf-8")

            with self.rtdb.begin_txn(self._funcsdb) as txn:
                value = txn.get(key)
                if value is None:
                    return None

                cmsg = function_pb2.Function()  # type: ignore  # pylint:disable=no-member
                cmsg.ParseFromString(value)

                # Reconstruct function
                func = Function.parse_from_cmessage(
                    cmsg,
                    function_manager=self._backref,
                    project=self._backref._kb._project,
                    meta_only=meta_only,
                )

            if meta_only:
                self._meta_func_cache[addr] = func
            else:
                # Remove from spilled set
                self._spilled_keys.discard(addr)
                # Add to in-memory map
                super().__setitem__(addr, func)
                self._on_function_stored(addr)

            # l.debug("Loaded function %s from LMDB", hex(addr) if isinstance(addr, int) else addr)
            return func
        finally:
            self._loading_from_lmdb = False

            # After loading is complete (and we're back to not loading), evict if needed
            if self._eviction_enabled and self._cache_limit is not None and self.cached_count > self._cache_limit:
                self._evict_lru()

    def load_all_spilled(self) -> None:
        """
        Load all spilled functions back into memory (disables eviction temporarily).
        """
        if not self._spilled_keys:
            return

        # Temporarily disable eviction
        old_eviction_state = self._eviction_enabled
        self._eviction_enabled = False

        try:
            # Make a copy of spilled_addrs since _load_from_lmdb modifies it
            addrs_to_load = list(self._spilled_keys)
            for addr in addrs_to_load:
                self._load_from_lmdb(addr)
        finally:
            self._eviction_enabled = old_eviction_state

    def evict_all_cached(self) -> None:
        """
        Evict all cached functions to LMDB.
        """
        if self.cached_count == 0:
            return
        self._evict_n(self.cached_count)


class FunctionManager(Generic[K], KnowledgeBasePlugin, collections.abc.Mapping[K, Function]):
    """
    When cache_limit is set, the FunctionManager uses a SpillingFunctionDict
    that implements an LRU cache keeping only the most recently accessed N functions
    in memory, spilling others to an LMDB database on disk. This allows working with
    binaries that have more functions than can fit in memory.

    :param cache_limit: Maximum number of functions to keep in memory. None means unlimited (no eviction). Default is
                        None.
    """

    def __init__(self, kb: KnowledgeBase, cache_limit: int | None = None):
        super().__init__(kb=kb)
        self.function_address_types = self._kb._project.arch.function_address_types
        self.address_types = self._kb._project.arch.address_types

        if cache_limit is None and USE_SPILLING_FUNCTION_DICT:
            cache_limit = self.get_default_cache_limit()

        # Use SpillingFunctionDict when caching is enabled, otherwise plain FunctionDict
        if cache_limit is not None:
            self._function_map: FunctionDict[K] | SpillingFunctionDict[K] = SpillingFunctionDict(
                self, kb.rtdb, key_types=self.function_address_types, cache_limit=cache_limit
            )
        else:
            self._function_map = FunctionDict(self, key_types=self.function_address_types)

        self.function_addrs_set: set = set()
        self.callgraph: networkx.MultiDiGraph[int] = networkx.MultiDiGraph()

        # Registers used for passing arguments around
        self._arg_registers = kb._project.arch.argument_registers

        # local PLT dictionary cache
        self._rplt_cache_ranges: None | list[tuple[int, int]] = None
        self._rplt_cache: None | set[int] = None
        # local binary name cache: min_addr -> (max_addr, binary_name)
        self._binname_cache: None | SortedDict[int, tuple[int, str | None]] = None

        # non-returning functions cache
        self._non_returning_func_addrs: set[K] = set()
        self._unknown_returning_func_addrs: set[K] = set()
        # function number of blocks cache
        self._func_block_counts: dict[K, int] = {}
        # function name cache
        self._func_name_to_addrs: defaultdict[str, set[K]] = defaultdict(set)
        # historical function name cache
        self._old_func_name_to_addrs: defaultdict[str, set[K]] = defaultdict(set)

    def __setstate__(self, state):
        self.function_address_types = state["function_address_types"]
        self.address_types = state["address_types"]
        self._function_map = state["_function_map"]
        self.callgraph = state["callgraph"]

        # Reinitialize cache state
        self._rplt_cache_ranges = None
        self._rplt_cache = None
        self._binname_cache = None
        self._non_returning_func_addrs = set()
        self._unknown_returning_func_addrs = set()
        self._func_block_counts = {}
        self._func_name_to_addrs = defaultdict(set)
        self._old_func_name_to_addrs = defaultdict(set)
        self.function_addrs_set = set()

        self._function_map._backref = weakref.proxy(self)
        for func in self._function_map.values():
            func._function_manager = self
            if func.returning is None:
                self._unknown_returning_func_addrs.add(func.addr)
            elif func.returning is False:
                self._non_returning_func_addrs.add(func.addr)
            self._func_block_counts[func.addr] = len(func.block_addrs_set)
            self._func_name_to_addrs[func.name].add(func.addr)
            for old_name in func.previous_names:
                self._old_func_name_to_addrs[old_name].add(func.addr)
            self.function_addrs_set.add(func.addr)

    def set_kb(self, kb: KnowledgeBase):
        super().set_kb(kb)
        # If the unpickled function_map is a SpillingFunctionDict, set rtdb properly
        if isinstance(self._function_map, SpillingFunctionDict):
            self._function_map.rtdb = kb.rtdb

    def __getstate__(self):
        return {
            "function_address_types": self.function_address_types,
            "address_types": self.address_types,
            "_function_map": self._function_map,
            "callgraph": self.callgraph,
        }

    def copy(self):
        cache_limit = None
        if isinstance(self._function_map, SpillingFunctionDict):
            cache_limit = self._function_map.cache_limit

        fm = FunctionManager(self._kb, cache_limit=cache_limit)
        for k, v in self._function_map.items():
            # Note that we use shallow copy of Function instances and do not update Function._function_manager to point
            # to fm.
            fm._function_map[k] = v
        fm._function_map._backref = weakref.proxy(fm)
        fm.callgraph = networkx.MultiDiGraph(self.callgraph)
        fm._arg_registers = self._arg_registers.copy()
        fm.function_addrs_set = self.function_addrs_set.copy()

        # cache
        fm._non_returning_func_addrs = self._non_returning_func_addrs.copy()
        fm._unknown_returning_func_addrs = self._unknown_returning_func_addrs.copy()
        fm._func_block_counts = self._func_block_counts.copy()
        fm._func_name_to_addrs = defaultdict(set, {k: v.copy() for k, v in self._func_name_to_addrs.items()})
        fm._old_func_name_to_addrs = defaultdict(set, {k: v.copy() for k, v in self._old_func_name_to_addrs.items()})

        return fm

    def clear(self):
        if isinstance(self._function_map, SpillingFunctionDict):
            cache_limit = self._function_map.cache_limit
            self._function_map.clear()
            self._function_map = SpillingFunctionDict(
                self, self._kb.rtdb, key_types=self.function_address_types, cache_limit=cache_limit
            )
        else:
            self._function_map = FunctionDict(self, key_types=self.function_address_types)

        self.callgraph = networkx.MultiDiGraph()
        self.function_addrs_set = set()
        # cache
        self._rplt_cache = None
        self._rplt_cache_ranges = None
        self._binname_cache = None
        self._non_returning_func_addrs = set()
        self._unknown_returning_func_addrs = set()
        self._func_block_counts = {}
        self._func_name_to_addrs = defaultdict(set)
        self._old_func_name_to_addrs = defaultdict(set)

    def get_default_cache_limit(self, max_limit: int = 5000) -> int | None:
        """
        Get the default function cache limit based on the size of the binary.

        :return: The default cache limit; None means unlimited.
        """
        if self._kb is None or self._kb._project is None:
            return max_limit
        limit = self._kb._project.get_function_cache_limit()
        if limit is None:
            return limit
        limit = max(limit, 100)
        return min(max_limit, limit)

    def _generate_callmap_sif(self, filepath):
        """
        Generate a sif file from the call map.

        :param filepath:    Path of the sif file
        :return:            None
        """
        with open(filepath, "w", encoding="utf-8") as f:
            f.writelines(f"{src:#x}\tDirectEdge\t{dst:#x}\n" for src, dst in self.callgraph.edges())

    def _addr_in_plt_cached_ranges(self, addr: int) -> bool:
        if self._rplt_cache_ranges is None:
            return False
        pos = bisect.bisect_left(self._rplt_cache_ranges, addr, key=lambda x: x[0])
        return pos > 0 and self._rplt_cache_ranges[pos - 1][0] <= addr < self._rplt_cache_ranges[pos - 1][1]

    def is_plt_cached(self, addr: int) -> bool:
        # check if the addr is in the cache range
        if not self._addr_in_plt_cached_ranges(addr):
            # find the object containing this addr
            obj = self._kb._project.loader.find_object_containing(addr, membership_check=False)
            if obj is None:
                return False
            if self._rplt_cache_ranges is None:
                self._rplt_cache_ranges = []
            obj_range = obj.min_addr, obj.max_addr
            idx = bisect.bisect_left(self._rplt_cache_ranges, obj_range)
            if not (idx < len(self._rplt_cache_ranges) and self._rplt_cache_ranges[idx] == obj_range):
                self._rplt_cache_ranges.insert(idx, obj_range)
            if isinstance(obj, (cle.MetaELF, cle.MachO)):
                if self._rplt_cache is None:
                    self._rplt_cache = set()
                self._rplt_cache |= set(obj.reverse_plt)

        return addr in self._rplt_cache if self._rplt_cache is not None else False

    def _binname_cache_get_addr_base(self, addr: int) -> int | None:
        if self._binname_cache is None:
            return None
        try:
            base_addr = next(self._binname_cache.irange(maximum=addr, reverse=True))
        except StopIteration:
            return None
        return base_addr if base_addr <= addr < self._binname_cache[base_addr][0] else None

    def get_binary_name_cached(self, addr: int) -> str | None:
        base_addr = self._binname_cache_get_addr_base(addr)
        if base_addr is None:
            # not cached; cache it first
            obj = self._kb._project.loader.find_object_containing(addr, membership_check=False)
            if obj is None:
                return None
            if self._binname_cache is None:
                self._binname_cache = SortedDict()
            binary_basename = os.path.basename(obj.binary) if obj.binary else None
            self._binname_cache[obj.min_addr] = obj.max_addr, binary_basename
            base_addr = obj.min_addr
        return self._binname_cache[base_addr][1] if self._binname_cache is not None else None

    def function_name_changed(self, addr: K, old_name: str | None, new_name: str) -> None:
        """
        Notify the FunctionManager that a function's name has changed.

        :param addr:        Address of the function.
        :param old_name:    Old name of the function, or None if there is no old name.
        :param new_name:    New name of the function.
        """
        if old_name is not None:
            if old_name in self._func_name_to_addrs:
                self._func_name_to_addrs[old_name].discard(addr)
                if not self._func_name_to_addrs[old_name]:
                    del self._func_name_to_addrs[old_name]
            self._old_func_name_to_addrs[old_name].add(addr)
        self._func_name_to_addrs[new_name].add(addr)

    def _add_node(self, function_addr, node, syscall=None, size=None):
        if isinstance(node, self.address_types):
            node = self._kb._project.factory.snippet(node, size=size)
        dst_func = self._function_map[function_addr]
        if syscall in (True, False):
            dst_func.is_syscall = syscall
        dst_func._register_node(True, node)

    def _add_call_to(
        self,
        function_addr,
        from_node,
        to_addr,
        retn_node=None,
        syscall: bool = False,
        stmt_idx=None,
        ins_addr=None,
        return_to_outside: bool = False,
    ):
        """
        Add a call to a function.

        :param int function_addr:   Address of the current function where this call happens.
        :param from_node:           The source node.
        :param to_addr:             Address of the target function, or None if unknown.
        :param retn_node:           The node where the target function will return to if it returns.
        :param bool syscall:        If this is a call to a syscall or not.
        :param int stmt_idx:        ID of the statement where this call happens.
        :param int ins_addr:        Address of the instruction where this call happens.
        :param return_to_outside:  True if the return of the call is considered going to outside of the current
                                        function.
        :return:                    None
        """

        if isinstance(from_node, self.address_types):
            from_node = self._kb._project.factory.snippet(from_node)
        if isinstance(retn_node, self.address_types):
            retn_node = self._kb._project.factory.snippet(retn_node)
        func = self._function_map[function_addr]
        func._add_call_site(from_node.addr, to_addr, retn_node.addr if retn_node else None)

        if to_addr is not None:
            self.function(addr=to_addr, create=True, syscall=syscall)
            dest_func_node = FuncNode(to_addr)
            func._call_to(
                from_node,
                dest_func_node,
                retn_node,
                stmt_idx=stmt_idx,
                ins_addr=ins_addr,
                return_to_outside=return_to_outside,
                syscall=syscall,
            )

        if return_to_outside:
            func.add_retout_site(from_node)

        # is there any existing edge on the callgraph?
        edge_data = {"type": "call"}
        if to_addr is not None and (
            function_addr not in self.callgraph
            or to_addr not in self.callgraph[function_addr]
            or edge_data not in self.callgraph[function_addr][to_addr].values()
        ):
            self.callgraph.add_edge(function_addr, to_addr, **edge_data)

    def _add_fakeret_to(
        self, function_addr, from_node, to_node, confirmed=None, syscall=None, to_outside=False, to_function_addr=None
    ):
        if isinstance(from_node, self.address_types):
            from_node = self._kb._project.factory.snippet(from_node)
        if isinstance(to_node, self.address_types):
            to_node = self._kb._project.factory.snippet(to_node)
        src_func = self._function_map[function_addr]

        if syscall in (True, False):
            src_func.is_syscall = syscall

        src_func._fakeret_to(from_node, to_node, confirmed=confirmed, to_outside=to_outside)

        if to_outside and to_function_addr is not None:
            # mark it on the callgraph
            edge_data = {"type": "fakeret"}
            if (
                function_addr not in self.callgraph
                or to_function_addr not in self.callgraph[function_addr]
                or edge_data not in self.callgraph[function_addr][to_function_addr].values()
            ):
                self.callgraph.add_edge(function_addr, to_function_addr, **edge_data)

    def _remove_fakeret(self, function_addr, from_node, to_node):
        if type(from_node) is int:  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        if type(to_node) is int:  # pylint: disable=unidiomatic-typecheck
            to_node = self._kb._project.factory.snippet(to_node)
        self._function_map[function_addr]._remove_fakeret(from_node, to_node)

    def _add_return_from(self, function_addr, from_node, to_node=None):  # pylint:disable=unused-argument
        if isinstance(from_node, self.address_types):  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        self._function_map[function_addr]._add_return_site(from_node)

    def _add_transition_to(self, function_addr, from_node, to_node, ins_addr=None, stmt_idx=None, is_exception=False):
        if isinstance(from_node, self.address_types):  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        if isinstance(to_node, self.address_types):  # pylint: disable=unidiomatic-typecheck
            to_node = self._kb._project.factory.snippet(to_node)
        self._function_map[function_addr]._transit_to(
            from_node, to_node, ins_addr=ins_addr, stmt_idx=stmt_idx, is_exception=is_exception
        )

    def _add_outside_transition_to(
        self, function_addr, from_node, to_node, to_function_addr=None, ins_addr=None, stmt_idx=None, is_exception=False
    ):
        if type(from_node) is int:  # pylint: disable=unidiomatic-typecheck
            from_node = self._kb._project.factory.snippet(from_node)
        if type(to_node) is int:  # pylint: disable=unidiomatic-typecheck
            try:
                to_node = self._kb._project.factory.snippet(to_node)
            except SimEngineError:
                # we cannot get the snippet, but we should at least tell the function that it's going to jump out here
                self._function_map[function_addr].add_jumpout_site(from_node)
                return
        self._function_map[function_addr]._transit_to(
            from_node,
            to_node,
            outside=True,
            ins_addr=ins_addr,
            stmt_idx=stmt_idx,
            is_exception=is_exception,
        )

        if to_function_addr is not None:
            # mark it on the callgraph
            edge_data = {"type": "transition" if not is_exception else "exception"}
            if (
                function_addr not in self.callgraph
                or to_function_addr not in self.callgraph[function_addr]
                or edge_data not in self.callgraph[function_addr][to_function_addr].values()
            ):
                self.callgraph.add_edge(function_addr, to_function_addr, **edge_data)

    def _add_return_from_call(self, function_addr, src_function_addr, to_node, to_outside=False):
        # Note that you will never return to a syscall

        if type(to_node) is int:  # pylint: disable=unidiomatic-typecheck
            to_node = self._kb._project.factory.snippet(to_node)
        func = self._function_map[function_addr]
        src_funcnode = FuncNode(src_function_addr)
        func._return_from_call(src_funcnode, to_node, to_outside=to_outside)

    #
    # Dict methods
    #

    def __contains__(self, item):
        if type(item) is int:
            # Delegate to _function_map (SpillingFunctionDict handles spilled addrs)
            return item in self._function_map

        try:
            _ = self[item]
            return True
        except (KeyError, TypeError):
            return False

    def __getitem__(self, k) -> Function:
        if isinstance(k, self.function_address_types):
            f = self.function(addr=k)
        elif type(k) is str:
            f = self.function(name=k) or self.function(name=k, check_previous_names=True)
        else:
            raise ValueError(f"FunctionManager.__getitem__ does not support keys of type {type(k)}")

        if f is None:
            raise KeyError(k)

        return f

    def __setitem__(self, k, v):
        if isinstance(k, self.function_address_types):
            self._function_map[k] = v
            self._function_added(v)
        else:
            raise ValueError("FunctionManager.__setitem__ keys must be an int")

    def __delitem__(self, k):
        if isinstance(k, self.function_address_types):
            func_meta = self._function_map.get(k, None, meta_only=True)
            if k in self._function_map:
                del self._function_map[k]
            if k in self.callgraph:
                self.callgraph.remove_node(k)
            self.function_addrs_set.discard(k)
            self._non_returning_func_addrs.discard(k)
            self._unknown_returning_func_addrs.discard(k)
            self._func_block_counts.pop(k, None)
            self._func_name_to_addrs.pop(k, None)
            if func_meta is not None:
                for old_name in func_meta.previous_names:
                    self._old_func_name_to_addrs.get(old_name, set()).discard(k)
        else:
            raise ValueError(
                f"FunctionManager.__delitem__ only accepts the following address types: {self.function_address_types}"
            )

    def __len__(self):
        # Delegate to _function_map (SpillingFunctionDict includes spilled count)
        return len(self._function_map)

    def __iter__(self):
        # Delegate to _function_map (SpillingFunctionDict handles spilled addrs)
        yield from self._function_map

    def items(self, /, meta_only: bool = False) -> Generator[tuple[K, Function]]:  # type: ignore
        for addr in self._function_map:
            yield addr, self._function_map.get(addr, meta_only=meta_only)

    def values(self, /, meta_only: bool = False) -> Generator[Function]:  # type: ignore
        for addr in self._function_map:
            yield self._function_map.get(addr, meta_only=meta_only)

    def get_by_addr(self, addr, meta_only: bool = False) -> Function:
        return self._function_map.get(addr, meta_only=meta_only)

    def get_by_name(self, name: str, check_previous_names: bool = False) -> Generator[Function]:
        if not check_previous_names:
            addrs = self._func_name_to_addrs.get(name, set())
            if all(isinstance(a, int) for a in addrs):
                addrs = sorted(addrs)
            for addr in addrs:
                yield self._function_map.get(addr)
        else:
            addrs = self._old_func_name_to_addrs.get(name, set()) | self._func_name_to_addrs.get(name, set())
            if all(isinstance(a, int) for a in addrs):
                addrs = sorted(addrs)
            for addr in addrs:
                yield self._function_map.get(addr)

    def get_addrs_by_name(self, name: str, check_previous_names: bool = False) -> set[int]:
        if not check_previous_names:
            return self._func_name_to_addrs.get(name, set()).copy()
        return self._old_func_name_to_addrs.get(name, set()) | self._func_name_to_addrs.get(name, set())

    def _function_added(self, func: Function):
        """
        A callback method for adding a new function instance to the manager.

        :param func:   The Function instance being added.
        :return:       None
        """

        # Add the function address to the set of function addresses
        self.function_addrs_set.add(func.addr)

        # update the non-returning cache
        self.set_function_returning(func.addr, func.returning)

        # update the function block count cache
        self.set_func_block_count(func.addr, len(func.block_addrs_set))

        # make sure all functions exist in the call graph
        self.callgraph.add_node(func.addr)

        # trigger function_name_changed to update function name caches
        self.function_name_changed(func.addr, None, func.name)

    def contains_addr(self, addr):
        """
        Decide if an address is handled by the function manager.

        Note: this function is non-conformant with python programming idioms, but its needed for performance reasons.

        :param int addr: Address of the function.
        """
        return addr in self._function_map

    def ceiling_addr(self, addr: K) -> K | None:
        """
        Return the function who has the least address that is greater than or equal to `addr`.

        :param int addr: The address to query.
        :return:         A Function instance, or None if there is no other function after `addr`.
        """

        try:
            return self._function_map.ceiling_addr(addr)
        except KeyError:
            return None

    def ceiling_func(self, addr) -> Function | None:
        """
        Return the function who has the least address that is greater than or equal to `addr`.

        :param int addr: The address to query.
        :return:         A Function instance, or None if there is no other function after `addr`.
        """

        try:
            next_addr = self._function_map.ceiling_addr(addr)
            return self._function_map.get(next_addr)

        except KeyError:
            return None

    def floor_addr(self, addr: K) -> K | None:
        """
        Return the function who has the greatest address that is less than or equal to `addr`.

        :param int addr: The address to query.
        :return:         An address, or None if there is no other function before `addr`.
        """

        try:
            return self._function_map.floor_addr(addr)
        except KeyError:
            return None

    def floor_func(self, addr):
        """
        Return the function who has the greatest address that is less than or equal to `addr`.

        :param int addr: The address to query.
        :return:         A Function instance, or None if there is no other function before `addr`.
        :rtype:          Function or None
        """

        try:
            prev_addr = self._function_map.floor_addr(addr)
            return self._function_map.get(prev_addr)

        except KeyError:
            return None

    def query(self, query: str, check_previous_names: bool = False) -> Function | None:
        """
        Query for a function using selectors to disambiguate. Supported variations:

            ::<name>           Function <name> in the main object
            ::<addr>::<name>   Function <name> at <addr>
            ::<obj>::<name>    Function <name> in <obj>

        """
        # FIXME: Proper mangle handling
        matches = QUERY_PATTERN.match(query)
        if matches:
            selector = matches.group(2)
            name = matches.group(3)

            if selector is not None and ADDR_PATTERN.fullmatch(selector):
                addr = cast(K, int(matches.group(2), 0))
                try:
                    func = self._function_map.get(addr)
                    if func.name == name or (check_previous_names and name in func.previous_names):
                        return func
                except KeyError:
                    pass

            obj_name = selector or self._kb._project.loader.main_object.binary_basename
            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                if func.binary_name == obj_name:
                    return func

        return None

    def function(
        self, addr=None, name=None, check_previous_names=False, create=False, syscall=False, plt=None
    ) -> Function | None:
        """
        Get a function object from the function manager.

        Pass either `addr` or `name` with the appropriate values.

        :param int addr: Address of the function.
        :param str name: Name of the function.
        :param bool create: Whether to create the function or not if the function does not exist.
        :param bool syscall: True to create the function as a syscall, False otherwise.
        :param bool or None plt: True to find the PLT stub, False to find a non-PLT stub, None to disable this
                                 restriction.
        :return: The Function instance, or None if the function is not found and create is False.
        :rtype: Function or None
        """
        if name is not None and name.startswith("sub_"):
            # first check if a function with the specified name exists
            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                if plt is None or func.is_plt == plt:
                    return func

            # then enter the syntactic sugar mode
            try:
                addr = cast(K, int(name.split("_")[-1], 16))
                name = None
            except ValueError:
                pass

        if addr is not None:
            try:
                f = self._function_map.get(addr)
                if plt is None or f.is_plt == plt:
                    return f
            except KeyError:
                if create:
                    # the function is not found
                    f = self._function_map[addr]
                    if name is not None:
                        f.name = name
                    if syscall:
                        f.is_syscall = True
                    return f
        elif name is not None:
            func = self.query(name, check_previous_names=check_previous_names)
            if func is not None:
                return func

            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                if plt is None or func.is_plt == plt:
                    return func

        return None

    def dbg_draw(self, prefix="dbg_function_"):
        for func_addr, func in self._function_map.items():
            filename = f"{prefix}{func_addr:#08x}.png"
            func.dbg_draw(filename)

    def rebuild_callgraph(self):
        self.callgraph = networkx.MultiDiGraph()
        for func_addr in self._function_map:
            self.callgraph.add_node(func_addr)
        for func in self._function_map.values():
            if func.block_addrs_set:
                for node in func.transition_graph:
                    if isinstance(node, (HookNode, FuncNode)) and self.contains_addr(node.addr):
                        self.callgraph.add_edge(func.addr, node.addr)
                    else:
                        inedges = func.transition_graph.in_edges(node, data=True)
                        for _, _, data in inedges:
                            if (
                                data.get("type") == "transition"
                                and data.get("outside") is True
                                and self.contains_addr(node.addr)
                            ):
                                self.callgraph.add_edge(func.addr, node.addr)

    #
    # Non-returning function cache
    #

    def set_function_returning(self, addr: K, v: bool | None) -> None:
        if v is False:
            self._non_returning_func_addrs.add(addr)
            self._unknown_returning_func_addrs.discard(addr)
        elif v is True:
            self._non_returning_func_addrs.discard(addr)
            self._unknown_returning_func_addrs.discard(addr)
        else:
            # v is None
            self._unknown_returning_func_addrs.add(addr)
            self._non_returning_func_addrs.discard(addr)

    def nonreturning_func_addrs(self) -> Generator[int]:
        """
        Yield all non-returning function addresses.
        """
        yield from self._non_returning_func_addrs

    def unknown_returning_func_addrs(self) -> Generator[int]:
        """
        Yield all function addresses with unknown returning status.
        """
        yield from self._unknown_returning_func_addrs

    def is_func_nonreturning(self, addr: K) -> bool:
        """
        Check if a function is non-returning.

        :param addr:    Address of the function.
        :return:        True if non-returning, False if returning or unknown.
        """
        return addr in self._non_returning_func_addrs

    def is_func_returning_unknown(self, addr: K) -> bool:
        """
        Check if a function's returning status is unknown.

        :param addr:    Address of the function.
        :return:        True if returning status is unknown, False otherwise.
        """
        return addr in self._unknown_returning_func_addrs

    #
    # Function block count cache
    #

    def get_func_block_count(self, addr: K) -> int | None:
        """
        Get the number of blocks in a function.

        :param addr:    Address of the function.
        :return:        Number of blocks, or None if unknown.
        """
        return self._func_block_counts.get(addr)

    def set_func_block_count(self, addr: K, count: int) -> None:
        """
        Set the number of blocks in a function.

        :param addr:    Address of the function.
        :param count:   Number of blocks.
        :return:        None
        """
        self._func_block_counts[addr] = count

    #
    # LRU Cache Management (delegates to SpillingFunctionDict when available)
    #

    @property
    def cache_limit(self) -> int | None:
        """
        Get the maximum number of functions to keep in memory. None means unlimited (no eviction).
        """
        if isinstance(self._function_map, SpillingFunctionDict):
            return self._function_map.cache_limit
        return None

    @cache_limit.setter
    def cache_limit(self, value: int) -> None:
        """
        Set the maximum number of functions to keep in memory.
        If the new limit is lower than the current number of cached functions, excess functions will be evicted to
        storage.
        """
        if isinstance(self._function_map, SpillingFunctionDict):
            self._function_map.cache_limit = value
        else:
            # Need to convert FunctionDict to SpillingFunctionDict
            old_map = self._function_map
            new_map = SpillingFunctionDict(
                self, self._kb.rtdb, key_types=self.function_address_types, cache_limit=value
            )
            # Disable eviction during bulk copy
            new_map._eviction_enabled = False
            # Copy existing functions
            for addr, func in old_map.items():
                new_map[addr] = func
            # Re-enable eviction
            new_map._eviction_enabled = True
            self._function_map = new_map
            # Now trigger eviction if needed
            self._function_map._evict_lru()

    @property
    def cached_function_count(self) -> int:
        """
        Return the number of functions currently in memory.
        """
        if isinstance(self._function_map, SpillingFunctionDict):
            return self._function_map.cached_count
        return len(self._function_map)

    @property
    def spilled_function_count(self) -> int:
        """
        Return the number of functions currently spilled to LMDB.
        """
        if isinstance(self._function_map, SpillingFunctionDict):
            return self._function_map.spilled_count
        return 0

    @property
    def total_function_count(self) -> int:
        """
        Return the total number of functions (in memory + spilled).
        """
        if isinstance(self._function_map, SpillingFunctionDict):
            return self._function_map.total_count
        return len(self._function_map)

    @property
    def _spilled_addrs(self) -> set:
        """
        Proxy for backward compatibility - access spilled addresses from SpillingFunctionDict.
        """
        if isinstance(self._function_map, SpillingFunctionDict):
            return self._function_map._spilled_keys
        return set()


KnowledgeBasePlugin.register_default("functions", FunctionManager)



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations
from angr.ailment.expression import BinaryOp, Convert, Insert, Const

from .base import PeepholeOptimizationExprBase


class RemoveConstInsert(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Insert(c0, c1, v) ==> (c0 & mask) | (v << c1)"
    expr_classes = (Insert,)

    def optimize(self, expr: Insert, **kwargs):
        if not (
            isinstance(expr.base, Const)
            and not expr.base.tags.get("uninitialized", False)
            and isinstance(expr.base.value, int)
            and isinstance(expr.offset, Const)
            and isinstance(expr.offset.value, int)
        ):
            return None

        # TODO fix big-endian?
        assert self.project is not None
        base = expr.base.value & ~((1 << expr.value.bits) - 1) << (expr.offset.value * self.project.arch.byte_width)
        value = Convert(None, expr.value.bits, expr.bits, False, expr.value)
        shifted = (
            BinaryOp(
                None,
                "Shl",
                [value, Const(None, None, expr.offset.value * self.project.arch.byte_width, expr.bits)],
                signed=False,
            )
            if expr.offset.value != 0
            else value
        )
        return BinaryOp(expr.idx, "Or", [shifted, Const(None, None, base, shifted.bits)], signed=False)



"""icicle.py: An angr engine that uses Icicle to execute code."""

from __future__ import annotations

import logging
import os
from contextlib import suppress
from dataclasses import dataclass
from typing import cast

import pypcode
from archinfo import Arch, ArchARMCortexM, ArchPcode, Endness
from typing_extensions import override

from angr.errors import SimMemoryError
from angr.engines.concrete import ConcreteEngine, HeavyConcreteState
from angr.engines.failure import SimEngineFailure
from angr.engines.hook import HooksMixin
from angr.engines.syscall import SimEngineSyscall
from angr.rustylib.icicle import ExceptionCode, Icicle, VmExit
from angr.state_plugins.edge_hitmap import SimStateEdgeHitmap

log = logging.getLogger(__name__)


PROCESSORS_DIR = os.path.join(os.path.dirname(pypcode.__file__), "processors")


@dataclass
class IcicleStateTranslationData:
    """
    Represents the saved information needed to convert an Icicle state back
    to an angr state.
    """

    base_state: HeavyConcreteState
    registers: set[str]
    mapped_pages: set[int]
    writable_pages: set[int]
    explicit_page_metadata: dict[int, int | None]
    initial_cpu_icount: int
    icicle_arch: str


class IcicleEngine(ConcreteEngine):
    """
    An angr engine that uses Icicle to execute concrete states. The purpose of
    this implementation is to provide a high-performance concrete execution
    engine in angr. While historically, angr has focused on symbolic execution,
    better support for concrete execution enables new use cases such as fuzzing
    in angr. This is ideal for testing bespoke binary targets, such as
    microcontroller firmware, which may be difficult to correctly harness for
    use with traditional fuzzing engines.

    This class is the base class for the Icicle engine. It implements execution
    by creating an Icicle instance, copying the state from angr to Icicle, and then
    running the Icicle instance. The results are then copied back to the angr
    state. When snapshot mode is enabled, the Icicle instance is reused across
    multiple runs, restoring from a snapshot and only syncing changed state.

    For a more complete implementation, use the UberIcicleEngine class, which
    intends to provide a more complete set of features, such as hooks and syscalls.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_emu: Icicle | None = None
        self._base_translation_data: IcicleStateTranslationData | None = None
        self._snapshot_mode: bool = False

    @staticmethod
    def __make_icicle_arch(arch: Arch) -> str | None:
        """
        Convert an angr architecture to an Icicle architecture. Not particularly
        accurate, just a set of heuristics to get the right architecture. When
        adding a new architecture, this function may need to be updated.
        """
        if isinstance(arch, ArchARMCortexM) or (isinstance(arch, ArchPcode) and arch.pcode_arch == "ARM:LE:32:Cortex"):
            return "armv7m"
        if arch.linux_name == "arm":
            return "armv7a" if arch.memory_endness == Endness.LE else "armeb"
        return arch.linux_name

    @staticmethod
    def __is_arm(icicle_arch: str) -> bool:
        """
        Check if the architecture is arm based on the address.
        """
        return icicle_arch.startswith(("arm", "thumb"))

    @staticmethod
    def __is_cortex_m(angr_arch: Arch, icicle_arch: str) -> bool:
        """
        Check if the architecture is cortex-m based on the address.
        """
        return isinstance(angr_arch, ArchARMCortexM) or icicle_arch == "armv7m"

    @staticmethod
    def __is_thumb(angr_arch: Arch, icicle_arch: str, addr: int) -> bool:
        """
        Check if the architecture is thumb based on the address.
        """
        return IcicleEngine.__is_cortex_m(angr_arch, icicle_arch) or (
            IcicleEngine.__is_arm(icicle_arch) and addr & 1 == 1
        )

    @staticmethod
    def __get_pages(state: HeavyConcreteState) -> set[int]:
        """
        Unfortunately, the memory model doesn't have a way to get all pages.
        Instead, we can get all of the backers from the loader, then all of the
        pages from the PagedMemoryMixin and then do some math.
        """
        pages = set()
        page_size = state.memory.page_size

        # pages from loader segments
        proj = state.project
        if proj is not None:
            for addr, backer in proj.loader.memory.backers():
                start = addr // page_size
                end = (addr + len(backer) - 1) // page_size
                pages.update(range(start, end + 1))

        # The paged memory model stores explicit page overrides in _pages.
        # A None entry means the page was explicitly unmapped and must shadow
        # loader backers.
        for page_num, page in state.memory._pages.items():
            if page is None:
                pages.discard(page_num)
            else:
                pages.add(page_num)

        return pages

    @staticmethod
    def __get_explicit_page_metadata(state: HeavyConcreteState) -> dict[int, int | None]:
        """
        Return explicit page overrides from the paged memory model.
        The key is page number. Value is permission bits, or None when the page
        is explicitly unmapped.
        """
        metadata = {}
        page_size = state.memory.page_size
        for page_num, page in state.memory._pages.items():
            if page is None:
                metadata[page_num] = None
            else:
                metadata[page_num] = state.memory.permissions(page_num * page_size).concrete_value
        return metadata

    @staticmethod
    def __convert_angr_state_to_icicle(state: HeavyConcreteState) -> tuple[Icicle, IcicleStateTranslationData]:
        icicle_arch = IcicleEngine.__make_icicle_arch(state.arch)
        if icicle_arch is None:
            raise ValueError("Unsupported architecture")

        proj = state.project
        if proj is None:
            raise ValueError("IcicleEngine requires a project to be set")

        emu = Icicle(icicle_arch, PROCESSORS_DIR, True, True)

        copied_registers = set()
        explicit_page_metadata = IcicleEngine.__get_explicit_page_metadata(state)

        # To create a state in Icicle, we need to do the following:
        # 1. Copy the register values
        for register in state.arch.register_list:
            register = register.vex_name.lower() if register.vex_name is not None else register.name
            try:
                emu.reg_write(
                    register,
                    state.solver.eval(state.registers.load(register), cast_to=int),
                )
                copied_registers.add(register)
            except KeyError:
                log.debug("Register %s not found in icicle", register)

        # Unset the thumb bit if necessary
        if IcicleEngine.__is_thumb(state.arch, icicle_arch, state.addr):
            emu.pc = state.addr & ~1
            emu.isa_mode = 1
        elif "arm" in icicle_arch:  # Hack to work around us calling it r15t
            emu.pc = state.addr

        # Special case for x86 gs register
        if state.arch.name == "X86":
            emu.reg_write("GS_OFFSET", state.registers.load("gs").concrete_value << 16)

        # 2. Copy the memory contents

        mapped_pages = IcicleEngine.__get_pages(state)
        writable_pages = set()
        for page_num in mapped_pages:
            addr = page_num * state.memory.page_size
            size = state.memory.page_size
            perm_bits = state.memory.permissions(addr).concrete_value
            emu.mem_map(addr, size, perm_bits)
            memory = state.memory.concrete_load(addr, size)
            emu.mem_write(addr, memory)

            if perm_bits & 2:
                writable_pages.add(page_num)

        # Add breakpoints for simprocedures
        for addr in proj._sim_procedures:
            emu.add_breakpoint(addr)

        translation_data = IcicleStateTranslationData(
            base_state=state,
            registers=copied_registers,
            mapped_pages=mapped_pages,
            writable_pages=writable_pages,
            explicit_page_metadata=explicit_page_metadata,
            initial_cpu_icount=emu.cpu_icount,
            icicle_arch=icicle_arch,
        )

        # 3. Copy edge hitmap
        if state.has_plugin("edge_hitmap"):
            hitmap_plugin = cast(SimStateEdgeHitmap, state.get_plugin("edge_hitmap"))
            if hitmap_plugin.edge_hitmap is not None:
                emu.edge_hitmap = hitmap_plugin.edge_hitmap

        return (emu, translation_data)

    @staticmethod
    def __convert_icicle_state_to_angr(
        emu: Icicle, translation_data: IcicleStateTranslationData, status: VmExit
    ) -> HeavyConcreteState:
        state = translation_data.base_state.copy()

        # 1. Copy the register values
        for register in translation_data.registers:
            state.registers.store(register, emu.reg_read(register))

        if IcicleEngine.__is_arm(emu.architecture):  # Hack to work around us calling it r15t
            state.registers.store("pc", (emu.pc | 1) if emu.isa_mode == 1 else emu.pc)

        # 2. Copy only memory pages that were actually modified during execution
        modified_addrs = set(emu.modified_pages)
        page_size = state.memory.page_size
        for page_num in translation_data.writable_pages:
            addr = page_num * page_size
            if addr in modified_addrs:
                state.memory.store(addr, emu.mem_read(addr, page_size))

        # 3. Set history
        # 3.1 history.jumpkind
        exc = emu.exception_code
        if status == VmExit.UnhandledException:
            if exc in (
                ExceptionCode.ReadUnmapped,
                ExceptionCode.ReadPerm,
                ExceptionCode.WriteUnmapped,
                ExceptionCode.WritePerm,
                ExceptionCode.ExecViolation,
            ):
                state.history.jumpkind = "Ijk_SigSEGV"
            elif exc == ExceptionCode.Syscall:
                state.history.jumpkind = "Ijk_Syscall"
            elif exc == ExceptionCode.Halt:
                state.history.jumpkind = "Ijk_Exit"
            elif exc == ExceptionCode.InvalidInstruction:
                state.history.jumpkind = "Ijk_NoDecode"
            else:
                state.history.jumpkind = "Ijk_EmFail"
        else:
            state.history.jumpkind = "Ijk_Boring"

        # 3.2 history.recent_bbl_addrs
        # Skip the last block, because it will be added by Successors
        state.history.recent_bbl_addrs.extend([b[0] for b in emu.recent_blocks][:-1])

        # 3.3. Set history.recent_instruction_count
        state.history.recent_instruction_count = emu.cpu_icount - translation_data.initial_cpu_icount

        # 3.4. Set edge hitmap in dedicated plugin if present
        if state.has_plugin("edge_hitmap"):
            hitmap_plugin = cast(SimStateEdgeHitmap, state.get_plugin("edge_hitmap"))
            hitmap_plugin.edge_hitmap = emu.edge_hitmap

        return state

    @staticmethod
    def __sync_angr_state_to_icicle(
        emu: Icicle,
        state: HeavyConcreteState,
        base_translation_data: IcicleStateTranslationData,
    ) -> IcicleStateTranslationData:
        """
        Sync only registers and writable pages from an angr state to an existing
        Icicle VM. This is much faster than a full conversion since it skips VM
        creation, memory mapping, and read-only page copies.
        """
        icicle_arch = base_translation_data.icicle_arch

        # 1. Copy register values
        for register in base_translation_data.registers:
            with suppress(KeyError):
                emu.reg_write(
                    register,
                    state.solver.eval(state.registers.load(register), cast_to=int),
                )

        # Handle thumb/ARM mode
        if IcicleEngine.__is_thumb(state.arch, icicle_arch, state.addr):
            emu.pc = state.addr & ~1
            emu.isa_mode = 1
        elif "arm" in icicle_arch:
            emu.pc = state.addr

        # Special case for x86 gs register
        if state.arch.name == "X86":
            emu.reg_write("GS_OFFSET", state.registers.load("gs").concrete_value << 16)

        # 2. Sync only mapping/permission deltas from explicit page changes.
        page_size = state.memory.page_size
        explicit_page_metadata = IcicleEngine.__get_explicit_page_metadata(state)
        base_explicit_page_metadata = base_translation_data.explicit_page_metadata

        candidate_pages = set(base_explicit_page_metadata).symmetric_difference(explicit_page_metadata)
        for page_num in set(base_explicit_page_metadata).intersection(explicit_page_metadata):
            if base_explicit_page_metadata[page_num] != explicit_page_metadata[page_num]:
                candidate_pages.add(page_num)

        mapped_pages = set(base_translation_data.mapped_pages)
        writable_pages = set()
        writable_pages.update(base_translation_data.writable_pages)

        for page_num in candidate_pages:
            addr = page_num * page_size
            old_mapped = page_num in mapped_pages

            try:
                perm_bits = state.memory.permissions(addr).concrete_value
                new_mapped = True
            except SimMemoryError:
                perm_bits = 0
                new_mapped = False

            if old_mapped and not new_mapped:
                emu.mem_unmap(addr, page_size)
                mapped_pages.remove(page_num)
                writable_pages.discard(page_num)
                continue

            if not old_mapped and new_mapped:
                emu.mem_map(addr, page_size, perm_bits)
                mapped_pages.add(page_num)
            elif old_mapped and new_mapped:
                base_perm_bits = base_translation_data.base_state.memory.permissions(addr).concrete_value
                if base_perm_bits != perm_bits:
                    emu.mem_protect(addr, page_size, perm_bits)

            if perm_bits & 2:
                writable_pages.add(page_num)
            else:
                writable_pages.discard(page_num)

        # 3. Copy writable page contents.
        for page_num in writable_pages:
            addr = page_num * page_size
            memory = state.memory.concrete_load(addr, page_size)
            emu.mem_write(addr, memory)

        return IcicleStateTranslationData(
            base_state=state,
            registers=base_translation_data.registers,
            mapped_pages=mapped_pages,
            writable_pages=writable_pages,
            explicit_page_metadata=explicit_page_metadata,
            initial_cpu_icount=emu.cpu_icount,
            icicle_arch=icicle_arch,
        )

    def enable_snapshot_mode(self) -> None:
        """Enable snapshot mode for VM reuse across multiple runs."""
        self._snapshot_mode = True

    def has_snapshot(self) -> bool:
        """Check if a snapshot is available for fast restore."""
        return self._cached_emu is not None and self._cached_emu.has_snapshot()

    def restore_and_sync(self, state: HeavyConcreteState) -> None:
        """Restore the cached VM from snapshot and sync the given state to it."""
        if self._cached_emu is None or self._base_translation_data is None:
            raise ValueError("No cached emulator. Run process_concrete first with snapshot mode enabled.")
        self._cached_emu.restore_snapshot()
        self._base_translation_data = self.__sync_angr_state_to_icicle(
            self._cached_emu, state, self._base_translation_data
        )

    @override
    def process_concrete(
        self,
        state: HeavyConcreteState,
        num_inst: int | None = None,
        extra_stop_points: set[int] | None = None,
    ) -> HeavyConcreteState:
        if self._cached_emu is not None and self._snapshot_mode:
            # Fast path: restore snapshot and sync only changed state
            if self._base_translation_data is None:
                raise ValueError("No base translation data. Run process_concrete first with snapshot mode enabled.")
            self._cached_emu.restore_snapshot()
            translation_data = self.__sync_angr_state_to_icicle(self._cached_emu, state, self._base_translation_data)
            emu = self._cached_emu
        else:
            # Full init path
            emu, translation_data = self.__convert_angr_state_to_icicle(state)
            if self._snapshot_mode and self._cached_emu is None:
                emu.save_snapshot()
                self._cached_emu = emu
                self._base_translation_data = translation_data

        # Set extra stop points, skip the current PC. Track which ones were
        # actually added so we can clean them up after the run.
        added_breakpoints = []
        if extra_stop_points is not None:
            for addr in extra_stop_points:
                addr = addr & ~1  # Clear thumb bit if set
                if emu.pc != addr and emu.add_breakpoint(addr):
                    added_breakpoints.append(addr)

        # Set the instruction count limit
        if num_inst is not None and num_inst > 0:
            emu.icount_limit = num_inst

        # Reset page modification tracking so only writes during execution
        # are recorded.  This clears per-page modified flags and the global
        # modified set, allowing __convert_icicle_state_to_angr to skip
        # pages that were not touched.
        page_size = state.memory.page_size
        emu.reset_page_modification_tracking([page_num * page_size for page_num in translation_data.writable_pages])

        # Run it
        status = emu.run()

        # Remove extra stop points to prevent accumulation across runs
        for addr in added_breakpoints:
            emu.remove_breakpoint(addr)

        return IcicleEngine.__convert_icicle_state_to_angr(emu, translation_data, status)


class UberIcicleEngine(SimEngineFailure, SimEngineSyscall, HooksMixin, IcicleEngine):
    """
    An extension of the IcicleEngine that uses mixins to add support for
    syscalls and hooks. Most users will prefer to use this engine instead of the
    IcicleEngine directly.
    """



from __future__ import annotations
from typing import Generic, TypeVar, overload

import networkx
from angr.codenode import CodeNode
from angr.knowledge_plugins.functions.function import Function
from angr.utils.graph import compute_dominance_frontier, Dominators

T_co = TypeVar("T_co", covariant=True)


class DominanceFrontier(Generic[T_co]):
    """
    Computes the dominance frontier of all nodes in a function graph, and provides an easy-to-use interface for
    querying the frontier information.
    """

    @overload
    def __init__(
        self, func: Function, func_graph: networkx.DiGraph[T_co], entry: T_co, exception_edges: bool = False
    ): ...
    @overload
    def __init__(
        self: DominanceFrontier[CodeNode],
        func: Function,
        func_graph: networkx.DiGraph[CodeNode] | None = None,
        entry: CodeNode | None = None,
        exception_edges: bool = False,
    ): ...

    def __init__(
        self, func: Function, func_graph: networkx.DiGraph[T_co] | None = None, entry=None, exception_edges=False
    ):  # type: ignore
        self.function = func
        self.func_graph = func_graph
        self.entry = entry
        self._exception_edges = exception_edges

        self.frontiers = self._compute()

    def _get_graph(self) -> networkx.DiGraph[T_co]:
        if self.func_graph is not None:
            return self.func_graph
        return self.function.graph_ex(exception_edges=self._exception_edges)  # type: ignore

    def _compute(self):
        g = self._get_graph()

        if self.entry is not None:
            startpoint = self.entry
        else:
            if self.function.startpoint is None:
                # The function might be empty or is corrupted (maybe the object is created manually)
                raise TypeError(f"Startpoint of function {self.function!r} is None. Is this function empty?")

            if self.function.startpoint not in g:
                # find the actual start point
                startpoint = next((nn for nn in g if nn.addr == self.function.startpoint.addr), None)  # type: ignore
                if startpoint is None:
                    raise ValueError(
                        f"Cannot find the startpoint of function {self.function!r} in the given function graph."
                    )
            else:
                startpoint = self.function.startpoint

        # Compute the dominator tree
        doms: Dominators[T_co] = Dominators(g, startpoint)  # type: ignore

        # Compute the dominance frontier
        return compute_dominance_frontier(g, doms.dom)


def calculate_iterated_dominace_frontier_set(frontiers: dict, blocks: set) -> set:
    last_frontier: set | None = None
    while True:
        frontier = set()
        for b in blocks:
            if b in frontiers:
                frontier |= frontiers[b]
        if last_frontier is not None and last_frontier == frontier:
            break
        last_frontier = frontier
        blocks |= frontier
    return last_frontier



from __future__ import annotations
import logging

from capstone.x86_const import X86_OP_MEM, X86_REG_RIP

from angr.simos import SimWindows
from .resolver import IndirectJumpResolver

l = logging.getLogger(name=__name__)


class AMD64PeIatResolver(IndirectJumpResolver):
    """
    A timeless indirect call/jump resolver for IAT in amd64 PEs.
    """

    def __init__(self, project):
        super().__init__(project, timeless=True)

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        if not isinstance(self.project.simos, SimWindows):
            return False
        if jumpkind not in {"Ijk_Call", "Ijk_Boring"}:
            return False

        insns = self.project.factory.block(addr).capstone.insns
        if not insns:
            return False
        if not insns[-1].insn.operands:
            return False

        opnd = insns[-1].insn.operands[0]
        # Must be of the form: call qword ptr [0xABCD]
        return bool(opnd.type == X86_OP_MEM and opnd.mem.disp and opnd.mem.base == X86_REG_RIP and opnd.mem.index == 0)

    def resolve(self, cfg, addr, func_addr, block, jumpkind, func_graph_complete: bool = True, **kwargs):  # pylint:disable=unused-argument
        call_insn = self.project.factory.block(addr).capstone.insns[-1].insn
        addr = (call_insn.disp + call_insn.address + call_insn.size) & 0xFFFF_FFFF_FFFF_FFFF
        target = cfg._fast_memory_load_pointer(addr)
        if target is None:
            l.warning("Address %#x does not appear to be mapped", addr)
            return False, []

        if not self.project.is_hooked(target):
            return False, []

        dest = self.project.hooked_by(target)
        l.debug("Resolved target to %s", dest.display_name)
        return True, [target]



# pylint:disable=no-self-use,too-many-boolean-expressions
from __future__ import annotations
from angr.ailment.expression import BinaryOp, Const, Convert

from .base import PeepholeOptimizationExprBase
from .utils import get_expr_shift_left_amount


class RemoveRedundantShifts(PeepholeOptimizationExprBase):
    """
    Remove redundant bitshift operations.
    """

    __slots__ = ()

    NAME = "Remove redundant bitshifts"
    expr_classes = (BinaryOp,)  # all expressions are allowed

    def optimize(self, expr: BinaryOp, **kwargs):
        # (expr << N) >> N  ==> Convert((M-N)->M, Convert(M->(M-N), expr))
        if expr.op in ("Shr", "Sar") and isinstance(expr.operands[1], Const):
            expr_a = expr.operands[0]
            n0 = expr.operands[1].value
            if isinstance(expr_a, BinaryOp) and expr_a.op in {"Shl", "Mul"} and isinstance(expr_a.operands[1], Const):
                n1 = get_expr_shift_left_amount(expr_a)
                if n0 == n1:
                    inner_expr = expr_a.operands[0]
                    conv_inner_expr = Convert(
                        None,
                        expr_a.bits,
                        expr_a.bits - n0,
                        expr.op == "Sar",  # is_signed
                        inner_expr,
                        **expr.tags,
                    )
                    return Convert(
                        None,
                        expr_a.bits - n0,
                        expr.bits,
                        False,
                        conv_inner_expr,
                        **expr.tags,
                    )

        # expr << 0  ==>  expr
        # expr >> 0  ==>  expr
        if expr.op in {"Shl", "Shr", "Sar"} and isinstance(expr.operands[1], Const) and expr.operands[1].value == 0:
            return expr.operands[0]

        mask_hi32bits = 0xFFFFFFFF_00000000
        exp_32bits = 0x1_00000000
        if (
            expr.op == "Shr"
            and isinstance(expr.operands[1], Const)
            and expr.operands[1].value == 32
            and isinstance(expr.operands[0], BinaryOp)
            and expr.operands[0].op == "Or"
        ):
            for op0, op1 in [expr.operands[0].operands, expr.operands[0].operands[::-1]]:
                if (
                    isinstance(op1, Convert)
                    and op1.from_bits == 32
                    and op1.to_bits == 64
                    and op1.from_type == Convert.TYPE_INT
                    and op1.to_type == Convert.TYPE_INT
                    and isinstance(op0, BinaryOp)
                    and op0.op == "And"
                ):
                    # (expr<64-bits> & 0xffffffff_00000000) | Conv(32->64, expr<32-bits>)) >> 32  ==>
                    #     expr<64-bits> >> 32
                    inner_op0, inner_op1 = op0.operands
                    if isinstance(inner_op1, Const) and inner_op1.value == mask_hi32bits:
                        if (
                            isinstance(inner_op0, BinaryOp)
                            and isinstance(inner_op0.operands[1], Const)
                            and inner_op0.operands[1].value == exp_32bits
                        ):
                            return inner_op0.operands[0]
                        return BinaryOp(expr.idx, "Shr", [inner_op0, expr.operands[1]], expr.signed, **expr.tags)
                    return BinaryOp(expr.idx, "Shr", [op0, expr.operands[1]], expr.signed, **expr.tags)

            for op0, op1 in [expr.operands[0].operands, expr.operands[0].operands[::-1]]:
                # ((v11 & 0xffff_ffff | 10.0 * 0x1_00000000) >> 32)   ==>   10.0
                if (
                    isinstance(op1, BinaryOp)
                    and op1.op == "Mul"
                    and isinstance(op1.operands[1], Const)
                    and op1.operands[1].value == 0x1_0000_0000
                ):
                    if (
                        isinstance(op0, BinaryOp)
                        and op0.op == "And"
                        and isinstance(op0.operands[1], Const)
                        and op0.operands[1].value == 0xFFFF_FFFF
                    ):
                        return op1.operands[0]
                    if (
                        isinstance(op0, Convert)
                        and op0.from_bits == 32
                        and op0.to_bits == 64
                        and op0.from_type == Convert.TYPE_INT
                        and op0.to_type == Convert.TYPE_INT
                    ):
                        return op1.operands[0]

        return None



# pylint:disable=super-init-not-called
from __future__ import annotations
import os
from collections import defaultdict

import networkx

from angr.codenode import BlockNode
from .function import Function


class SootFunction(Function):
    """
    A representation of a function and various information about it.
    """

    def __init__(self, function_manager, addr, name=None, syscall=None):
        """
        Function constructor for Soot

        :param addr:            The address of the function.
        :param name:            (Optional) The name of the function.
        :param syscall:         (Optional) Whether this function is a syscall or not.
        """
        self.transition_graph = networkx.DiGraph()
        self._local_transition_graph = None
        # The Shimple CFG is already normalized.
        self.normalized = True

        # block nodes at whose ends the function returns
        self._ret_sites = set()
        # block nodes at whose ends the function jumps out to another function (jumps outside)
        self._jumpout_sites = set()
        # block nodes at whose ends the function calls out to another non-returning function
        self._callout_sites = set()
        # block nodes that ends the function by returning out to another function (returns outside). This is rare.
        self._retout_sites = set()
        # block nodes (basic block nodes) at whose ends the function terminates
        # in theory, if everything works fine, endpoints == ret_sites | jumpout_sites | callout_sites
        self._endpoints = defaultdict(set)
        self.previous_names = []

        self._call_sites = {}
        self.addr = addr
        self._function_manager = function_manager
        self._is_syscall = syscall

        self._project = project = self._function_manager._kb._project

        self._is_plt = False
        self._is_simprocedure = False

        if project.is_hooked(addr):
            self._is_simprocedure = True

        binary_name = None
        if self._is_simprocedure:
            hooker = project.hooked_by(addr)
            if hooker is not None:
                binary_name = hooker.library_name

        if binary_name is None and self.binary is not None:
            binary_name = os.path.basename(self.binary.binary)

        self._name = addr.__repr__()
        self.binary_name = binary_name

        # Stack offsets of those arguments passed in stack variables
        self._argument_stack_variables = []

        # These properties are set by VariableManager
        self.bp_on_stack = False
        self.retaddr_on_stack = False

        self.sp_delta = 0

        # Calling convention
        self.calling_convention = None

        # Function prototype
        self.prototype = None

        # Whether this function returns or not. `None` means it's not determined yet
        self._returning = None

        self._is_alignment = None

        # Determine returning status for SimProcedures and Syscalls
        hooker = None
        if self.is_simprocedure:
            hooker = project.hooked_by(addr)
        if hooker and hasattr(hooker, "NO_RET"):
            self.returning = not hooker.NO_RET

        # startpoint can always be None if this CFGNode is a syscall node
        self.startpoint = None

        self._addr_to_block_node = {}  # map addresses to nodes
        self._block_sizes = {}  # map addresses to block sizes
        self._block_cache = {}  # a cache of real, hard data Block objects
        self._local_blocks = {}  # a dict of all blocks inside the function
        self._local_block_addrs = set()  # a set of addresses of all blocks inside the function

        self.info = {}  # storing special information, like $gp values for MIPS32
        self.tags = ()  # store function tags. can be set manually by performing CodeTagging analysis.

    def normalize(self):
        # The Shimple CFG is already normalized.
        pass

    def _register_node(self, is_local: bool, node, update_func_block_count: bool = True):  # pylint:disable=unused-argument
        if is_local and self._local_blocks.get(node.addr) == node:
            return self._local_blocks[node.addr]

        if node not in self.transition_graph:
            self.transition_graph.add_node(node)
        node._graph = self.transition_graph
        if node.addr not in self or self._block_sizes[node.addr] == 0:
            self._block_sizes[node.addr] = node.size
        if node.addr == self.addr.addr and (self.startpoint is None or not self.startpoint.is_hook):
            self.startpoint = node
        if is_local:
            self._local_blocks[node.addr] = node
            self._local_block_addrs.add(node.addr)
        # add BlockNodes to the addr_to_block_node cache if not already there
        if isinstance(node, BlockNode) and node.addr not in self._addr_to_block_node:
            self._addr_to_block_node[node.addr] = node
        return node



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Generic, TypeVar
from typing_extensions import Self

import claripy

from angr.errors import SimMemoryError
from angr.state_plugins.plugin import SimStatePlugin

InData = TypeVar("InData")
OutData = TypeVar("OutData")
Addr = TypeVar("Addr")


class MemoryMixin(Generic[InData, OutData, Addr], SimStatePlugin):
    """
    MemoryMixin is the base class for the memory model in angr. It provides a
    set of methods that should be implemented by memory models. This is done
    using mixins, where each mixin handles some specific feature of the memory
    model, only overriding methods that it needs to implement its function. The
    memory model class itself then combines a set of mixins using inheritence to
    form the final memory model class.
    """

    # pylint: disable=no-self-use,unused-argument

    SUPPORTS_CONCRETE_LOAD: bool = False

    def __init__(self, memory_id: str | None = None, endness: str = "Iend_BE"):
        super().__init__()
        self.id = memory_id
        self.endness = endness

    @SimStatePlugin.memo
    def copy(self, memo):
        o = type(self).__new__(type(self))
        o.id = self.id
        o.endness = self.endness
        return o

    @property
    def category(self) -> str:
        """
        Return the category of this SimMemory instance. It can be one of the three following categories: reg, mem,
        or file.
        """

        assert isinstance(self.id, str)
        if self.id in ("reg", "mem"):
            return self.id

        if self.id.startswith("file"):
            return "file"

        if "_" in self.id:
            return self.id.split("_")[0]

        raise SimMemoryError(f'Unknown SimMemory category for memory_id "{self.id}"')

    @property
    def variable_key_prefix(self) -> tuple[Any, ...]:
        s = self.category
        if s == "file":
            return (s, self.id)
        return (s,)

    def find(
        self, addr: Addr, data: InData, max_search: int, **kwargs
    ) -> tuple[Addr, list[claripy.ast.Bool], list[int]]: ...

    def _add_constraints(self, c, *, add_constraints=True, condition=None, **kwargs):
        if add_constraints:
            to_add = c & condition | ~condition if condition is not None else c
            self.state.add_constraints(to_add)

    def load(self, addr: Addr, size: InData | None = None, **kwargs) -> OutData: ...

    def store(self, addr: Addr, data: InData, size: InData | None = None, **kwargs) -> None: ...

    def merge(
        self, others: list[Self], merge_conditions: list[claripy.ast.Bool] | None, common_ancestor: Self | None = None
    ) -> bool: ...

    def compare(self, other: Self) -> bool: ...

    def widen(self, others: list[Self]) -> bool: ...

    def permissions(self, addr: Addr, permissions: int | claripy.ast.BV | None = None, **kwargs) -> claripy.ast.BV: ...

    def map_region(
        self, addr: Addr, length: int, permissions: int | claripy.ast.BV, *, init_zero: bool = False, **kwargs
    ):
        pass

    def unmap_region(self, addr: Addr, length: int, **kwargs):
        pass

    # Optional interface:
    # TODO type this correctly. maybe split the bitmap version into another function?
    def concrete_load(self, addr, size, writing=False, **kwargs) -> Any:
        """
        Set SUPPORTS_CONCRETE_LOAD to True and implement concrete_load if reading concrete bytes is faster in this
        memory model.

        :param addr:    The address to load from.
        :param size:    Size of the memory read.
        :param writing:
        :return:        A memoryview into the loaded bytes.
        """
        raise NotImplementedError

    def erase(self, addr: Addr, size: int | None = None, **kwargs) -> None:
        """
        Set [addr:addr+size) to uninitialized. In many cases this will be faster than overwriting those locations with
        new values. This is commonly used during static data flow analysis.

        :param addr:    The address to start erasing.
        :param size:    The number of bytes for erasing.
        :return:        None
        """
        raise NotImplementedError

    def _default_value(self, addr, size, *, name=None, inspect=True, events=True, key=None, **kwargs) -> OutData:
        """
        Override this method to provide default values for a variety of edge cases and base cases.

        :param addr:    If this value is being filled to provide a default memory value, this will be its address.
                        Otherwise, None.
        :param size:    The size in bytes of the value to return
        :param name:    A descriptive identifier for the value, for if a symbol is created.

        The ``inspect``, ``events``, and ``key`` parameters are for ``state.solver.Unconstrained``, if it is used.
        """
        raise NotImplementedError

    def _merge_values(self, values: Iterable[tuple[Any, Any]], merged_size: int, **kwargs) -> OutData | None:
        """
        Override this method to provide value merging support.

        :param values:          A collection of values with their merge conditions.
        :param merged_size:     The size (in bytes) of the merged value.
        :return:                The merged value, or None to skip merging of the current value.
        """
        raise NotImplementedError

    def _merge_labels(self, labels: Iterable[dict], **kwargs) -> dict | None:
        """
        Override this method to provide label merging support.

        :param labels:          A collection of labels.
        :return:                The merged label, or None to skip merging of the current label.
        """
        raise NotImplementedError

    def replace_all(self, old: claripy.ast.BV, new: claripy.ast.BV):
        raise NotImplementedError

    def _replace_all(self, addrs: Iterable[int], old: claripy.ast.BV, new: claripy.ast.BV):
        raise NotImplementedError

    def copy_contents(self, dst: Addr, src: Addr, size: InData, condition: claripy.ast.Bool | None = None, **kwargs):
        """
        Override this method to provide faster copying of large chunks of data.

        :param dst:     The destination of copying.
        :param src:     The source of copying.
        :param size:    The size of copying.
        :param condition:   The storing condition.
        :param kwargs:      Other parameters.
        :return:        None
        """
        raise NotImplementedError



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from abc import abstractmethod
from typing import Any, TYPE_CHECKING
from collections import defaultdict
import weakref

from typing_extensions import Self

import angr.ailment as ailment
import claripy
import archinfo

from angr.errors import SimMemoryMissingError
from angr.project import Project
from angr.storage.memory_object import SimMemoryObject, SimLabeledMemoryObject
from angr.storage.memory_mixins import LabeledMemory
from angr.engines.light.engine import SimEngineLight
from angr.code_location import CodeLocation

if TYPE_CHECKING:
    from archinfo import Arch


class CallExprFinder(ailment.AILBlockRewriter):
    """
    Walks an AIL expression to find if it contains a call expression anywhere.
    """

    def __init__(self):
        super().__init__()
        self.has_call = False

    # pylint:disable=unused-argument
    def _handle_CallExpr(
        self,
        expr_idx: int,
        expr: ailment.Expr.Call,
        stmt_idx: int,
        stmt: ailment.Stmt.Statement,
        block: ailment.Block | None,
    ):
        self.has_call = True


class PropagatorState:
    """
    Describes the base state used in Propagator.

    :ivar arch:             Architecture of the binary.
    :ivar gp:               value of the global pointer for MIPS binaries.
    :ivar _replacements:    Stores expressions to replace, keyed by CodeLocation instances
    :ivar _only_consts:     Only track constants.
    :ivar _expr_used_locs:  A dict keyed by expressions and valued by CodeLocations where the expression is used.
    :ivar _max_prop_expr_occurrence:    The upperbound for the number of occurrences of an expression for Propagator
                            to propagate that expression to new locations (and replace the original expression).
                            Setting it to 0 disables this limit, which means Propagator will always propagate
                            expressions regardless of how many times it has been propagated.
    """

    __slots__ = (
        "__weakref__",
        "_artificial_reg_offsets",
        "_expr_used_locs",
        "_gp",
        "_max_prop_expr_occurrence",
        "_only_consts",
        "_replacements",
        "_store_tops",
        "arch",
        "gpr_size",
        "model",
        "project",
    )

    _tops = {}

    def __init__(
        self,
        arch: Arch,
        project: Project | None = None,
        replacements: defaultdict[CodeLocation, dict] | None = None,
        only_consts: bool = False,
        expr_used_locs: defaultdict[Any, set[CodeLocation]] | None = None,
        store_tops: bool = True,
        gp: int | None = None,
        max_prop_expr_occurrence: int = 1,
        model=None,
        artificial_reg_offsets=None,
    ):
        self.arch = arch
        self.gpr_size = arch.bits // arch.byte_width  # size of the general-purpose registers

        # propagation count of each expression
        self._expr_used_locs = defaultdict(list) if expr_used_locs is None else expr_used_locs
        self._only_consts = only_consts
        self._replacements = defaultdict(dict) if replacements is None else replacements
        self._store_tops = store_tops
        self._max_prop_expr_occurrence = max_prop_expr_occurrence
        self._artificial_reg_offsets = artificial_reg_offsets if artificial_reg_offsets is not None else set()

        # architecture-specific information
        self._gp: int | None = gp  # Value of gp for MIPS32 and 64 binaries

        self.project = project
        self.model = model

    def __repr__(self) -> str:
        return "<PropagatorState>"

    @classmethod
    def initial_state(
        cls,
        project: Project,
        only_consts=False,
        gp=None,
        do_binops=True,
        store_tops=False,
        func_addr=None,
        max_prop_expr_occurrence=None,
        initial_codeloc=None,
    ):
        raise NotImplementedError

    def _get_weakref(self):
        return weakref.proxy(self)

    @staticmethod
    def _is_const(v) -> bool:
        if isinstance(v, (int, ailment.Expr.Const)):
            return True
        if isinstance(v, claripy.ast.BV) and v.op == "BVV":
            return True
        if isinstance(v, claripy.ast.FP) and v.op == "FPV":
            return True
        return bool(isinstance(v, claripy.ast.Bool) and v.op == "BoolV")

    @staticmethod
    def _mo_cmp(
        mo_self: SimMemoryObject | SimLabeledMemoryObject,
        mo_other: SimMemoryObject | SimLabeledMemoryObject,
        addr: int,
        size: int,
    ):  # pylint:disable=unused-argument
        # comparing bytes from two sets of memory objects
        # we don't need to resort to byte-level comparison. object-level is good enough.

        # TODO what if object is bytes?
        if mo_self.object.symbolic or mo_other.object.symbolic:
            if type(mo_self) is SimLabeledMemoryObject and type(mo_other) is SimLabeledMemoryObject:
                return mo_self.label == mo_other.label and mo_self.object is mo_other.object
            if type(mo_self) is SimMemoryObject and type(mo_other) is SimMemoryObject:
                return mo_self.object is mo_other.object
            # SimMemoryObject vs SimLabeledMemoryObject -> the label must be different
            return False
        return None

    @staticmethod
    def top(bits: int) -> claripy.ast.BV:
        """
        Get a TOP value.

        :param size:    Width of the TOP value (in bits).
        :return:        The TOP value.
        """

        if bits in PropagatorState._tops:
            return PropagatorState._tops[bits]
        r = claripy.BVS("TOP", bits, explicit_name=True)
        PropagatorState._tops[bits] = r
        return r

    @staticmethod
    def is_top(expr) -> bool:
        """
        Check if the given expression is a TOP value.

        :param expr:    The given expression.
        :return:        True if the expression is TOP, False otherwise.
        """
        if isinstance(expr, claripy.ast.Base):
            if expr.op == "BVS" and expr.args[0] == "TOP":
                return True
            if "TOP" in expr.variables:
                return True
        return False

    @staticmethod
    def merge_replacements(replacements_0, replacements_1) -> bool:
        """
        The replacement merging logic is special: replacements_1 is the newer replacement result and replacement_0 is
        the older result waiting to be updated. When both replacements_1 and replacement_0 have a non-top value for the
        same variable and code location, we will update the slot in replacement_0 with the value from replacement_1.

        :return:            Whether merging has happened or not.
        """

        def _get_repl_size(repl_value: dict | ailment.Expression | claripy.ast.Bits) -> int:
            if isinstance(repl_value, dict):
                return _get_repl_size(repl_value["expr"])
            if isinstance(repl_value, ailment.Expression):
                return repl_value.bits
            return repl_value.size()

        merge_occurred = False
        for loc, vars_ in replacements_1.items():
            if loc not in replacements_0:
                replacements_0[loc] = vars_.copy()
                merge_occurred = True
            else:
                for var, repl in vars_.items():
                    if var not in replacements_0[loc]:
                        replacements_0[loc][var] = repl
                        merge_occurred = True
                    else:
                        if PropagatorState.is_top(repl) or PropagatorState.is_top(replacements_0[loc][var]):
                            if not PropagatorState.is_top(replacements_0[loc][var]):
                                t = PropagatorState.top(_get_repl_size(repl))
                                replacements_0[loc][var] = t
                                merge_occurred = True
                        elif (
                            (
                                isinstance(replacements_0[loc][var], claripy.ast.Base)
                                or isinstance(repl, claripy.ast.Base)
                            )
                            and replacements_0[loc][var] is not repl
                        ) or (
                            not isinstance(replacements_0[loc][var], claripy.ast.Base)
                            and not isinstance(repl, claripy.ast.Base)
                            and replacements_0[loc][var] != repl
                        ):
                            replacements_0[loc][var] = repl
                            merge_occurred = True
        return merge_occurred

    @abstractmethod
    def copy(self) -> Self:
        raise NotImplementedError

    def merge(self, *others: Self) -> tuple[Self, bool]:
        state = self.copy()
        merge_occurred = False

        for o in others:
            merge_occurred |= PropagatorState.merge_replacements(state._replacements, o._replacements)

        return state, merge_occurred

    def init_replacements(self):
        self._replacements = defaultdict(dict)

    def add_replacement(
        self,
        codeloc: CodeLocation,
        old,
        new,
        force_replace: bool = False,  # pylint:disable=unused-argument
    ) -> bool:
        """
        Add a replacement record: Replacing expression `old` with `new` at program location `codeloc`.
        If the self._only_consts flag is set to true, only constant values will be set.

        :param codeloc:                 The code location.
        :param old:                     The expression to be replaced.
        :param new:                     The expression to replace with.
        :return:                        True if the replacement will happen. False otherwise.
        """
        if self.is_top(new):
            return False

        replaced = False
        if self._only_consts:
            if self._is_const(new) or self.is_top(new):
                self._replacements[codeloc][old] = new
                replaced = True
        else:
            self._replacements[codeloc][old] = new
            replaced = True

        return replaced

    def filter_replacements(self):
        pass

    def has_replacements_at(self, codeloc: CodeLocation) -> bool:
        if not self._replacements:
            return False
        if codeloc not in self._replacements:
            return False
        return not all(self.is_top(replaced_by) for replaced_by in self._replacements[codeloc].values())


# VEX state


class RegisterAnnotation(claripy.Annotation):
    """
    Annotates TOP values that are coming from registers.
    """

    def __init__(self, offset, size):
        self.offset = offset
        self.size = size

    @property
    def eliminatable(self) -> bool:
        return True

    @property
    def relocatable(self) -> bool:
        return True

    def __hash__(self):
        return hash((RegisterAnnotation, self.offset, self.size))

    def __eq__(self, other):
        return type(other) is RegisterAnnotation and self.offset == other.offset and self.size == other.size


class RegisterComparisonAnnotation(claripy.Annotation):
    """
    Annotate TOP values that are the result of register values comparing against constant values.
    """

    def __init__(self, offset: int, size: int, cmp_op: str, value: int):
        self.offset = offset
        self.size = size
        self.cmp_op = cmp_op
        self.value = value

    @property
    def eliminatable(self) -> bool:
        return True

    @property
    def relocatable(self) -> bool:
        return True

    def __hash__(self):
        return hash((RegisterComparisonAnnotation, self.offset, self.size, self.cmp_op, self.value))

    def __eq__(self, other):
        return (
            type(other) is RegisterComparisonAnnotation
            and self.offset == other.offset
            and self.size == other.size
            and self.cmp_op == other.cmp_op
            and self.value == other.value
        )


class PropagatorVEXState(PropagatorState):
    """
    Describes the state used in the VEX engine of Propagator.
    """

    __slots__ = (
        "_registers",
        "_stack_variables",
        "block_initial_reg_values",
        "do_binops",
    )

    def __init__(
        self,
        arch,
        project=None,
        registers=None,
        local_variables=None,
        replacements=None,
        only_consts=False,
        expr_used_locs=None,
        do_binops=True,
        store_tops=True,
        block_initial_reg_values=None,
        gp=None,
        max_prop_expr_occurrence: int = 1,
        model=None,
        artificial_reg_offsets=None,
    ):
        super().__init__(
            arch,
            project=project,
            replacements=replacements,
            only_consts=only_consts,
            expr_used_locs=expr_used_locs,
            store_tops=store_tops,
            gp=gp,
            max_prop_expr_occurrence=max_prop_expr_occurrence,
            model=model,
            artificial_reg_offsets=artificial_reg_offsets,
        )
        self.do_binops = do_binops
        self._registers = (
            LabeledMemory(memory_id="reg", top_func=self.top, page_kwargs={"mo_cmp": self._mo_cmp})
            if registers is None
            else registers
        )
        self._stack_variables = (
            LabeledMemory(memory_id="mem", top_func=self.top, page_kwargs={"mo_cmp": self._mo_cmp})
            if local_variables is None
            else local_variables
        )

        self._registers.set_state(self)
        self._stack_variables.set_state(self)
        self.block_initial_reg_values = (
            defaultdict(list) if block_initial_reg_values is None else block_initial_reg_values
        )

    def __repr__(self):
        return "<PropagatorVEXState>"

    @classmethod
    def initial_state(
        cls,
        project,
        only_consts=False,
        gp=None,
        do_binops=True,
        store_tops=False,
        func_addr=None,
        max_prop_expr_occurrence=None,
        initial_codeloc=None,
        model=None,
    ):
        state = cls(
            project.arch,
            project=project,
            only_consts=only_consts,
            do_binops=do_binops,
            store_tops=store_tops,
            gp=gp,
            max_prop_expr_occurrence=max_prop_expr_occurrence,
            model=model,
        )
        spoffset_var = SimEngineLight.sp_offset(project.arch.bits, 0)
        state.store_register(
            project.arch.sp_offset,
            project.arch.bytes,
            spoffset_var,
        )
        if project.arch.name == "MIPS64":
            if func_addr is not None:
                state.store_register(  # pylint:disable=too-many-function-args
                    project.arch.registers["t9"][0],
                    project.arch.registers["t9"][1],
                    claripy.BVV(func_addr, 64),
                )
        elif project.arch.name == "MIPS32":
            if func_addr is not None:
                state.store_register(  # pylint:disable=too-many-function-args
                    project.arch.registers["t9"][0],
                    project.arch.registers["t9"][1],
                    claripy.BVV(func_addr, 32),
                )
        elif archinfo.arch_arm.is_arm_arch(project.arch):
            state.store_register(  # pylint:disable=too-many-function-args
                project.arch.registers["fpscr"][0],
                project.arch.registers["fpscr"][1],
                claripy.BVV(0, 32),
            )
        return state

    def copy(self) -> PropagatorVEXState:
        return PropagatorVEXState(
            self.arch,
            project=self.project,
            registers=self._registers.copy(),
            local_variables=self._stack_variables.copy(),
            replacements=self._replacements.copy(),
            expr_used_locs=self._expr_used_locs.copy(),
            only_consts=self._only_consts,
            do_binops=self.do_binops,
            store_tops=self._store_tops,
            block_initial_reg_values=self.block_initial_reg_values.copy(),
            gp=self._gp,
            max_prop_expr_occurrence=self._max_prop_expr_occurrence,
            model=self.model,
            artificial_reg_offsets=self._artificial_reg_offsets,
        )

    def merge(self, *others: PropagatorVEXState) -> tuple[PropagatorVEXState, bool]:
        state = self.copy()
        merge_occurred = state._registers.merge([o._registers for o in others], None)
        merge_occurred |= state._stack_variables.merge([o._stack_variables for o in others], None)
        return state, merge_occurred

    def store_local_variable(self, offset, size, value, endness):  # pylint:disable=unused-argument
        # TODO: Handle size
        self._stack_variables.store(offset, value, size=size, endness=endness)

    def load_local_variable(self, offset, size, endness) -> claripy.ast.BV:  # pylint:disable=unused-argument
        # TODO: Handle size
        try:
            return self._stack_variables.load(offset, size=size, endness=endness)
        except SimMemoryMissingError:
            return self.top(size * self.arch.byte_width)

    def store_register(self, offset, size, value):
        self._registers.store(offset, value, size=size)

    def load_register(self, offset, size):
        try:
            v = self._registers.load(offset, size=size)
            if self.is_top(v):
                v = v.annotate(RegisterAnnotation(offset, size))
            return v
        except SimMemoryMissingError:
            return self.top(size * self.arch.byte_width).annotate(RegisterAnnotation(offset, size))

    def register_results(self) -> dict[str, claripy.ast.BV]:
        result = {}
        for reg, (offset, size) in self.arch.registers.items():
            val = self.load_register(offset, size)
            if not self.is_top(val):
                result[reg] = val
        return result


# AIL state


class Equivalence:
    """
    Describes an equivalence relationship between two atoms.
    """

    __slots__ = (
        "atom0",
        "atom1",
        "codeloc",
        "is_weakassignment",
    )

    def __init__(self, codeloc, atom0, atom1, is_weakassignment: bool = False):
        self.codeloc = codeloc
        self.atom0 = atom0
        self.atom1 = atom1
        self.is_weakassignment = is_weakassignment

    def __repr__(self):
        return f"<Eq@{self.codeloc!r}: {self.atom0!r}=={self.atom1!r}>"

    def __eq__(self, other):
        return (
            type(other) is Equivalence
            and other.codeloc == self.codeloc
            and other.atom0 == self.atom0
            and other.atom1 == self.atom1
            and other.is_weakassignment == self.is_weakassignment
        )

    def __hash__(self):
        return hash((Equivalence, self.codeloc, self.atom0, self.atom1, self.is_weakassignment))
